/*--------------------------------------------------------------------------
 Mail Filter for CommuniGatePro
                                       ID Antivirus Lab http://www.drweb.ru
                                        http://www.drweb.ru/linux-bsd/forum
                                              http://www.drweb.ru/linux-bsd
----------------------------------------------------------------------------
 Acknowledgements:
 - Andy Igoshin <ai@vsu.ru>
--------------------------------------------------------------------------*/

#ifndef lint
static const char rcsid[] = "$Id: drweb_cgp.c,v 1.23 2001/10/16 15:28:21 cvser Exp $";
#endif

#if defined(__UNIX_DW)
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
#include <windows.h>
#include <process.h>
#include <io.h>
#define PATH_MAX    _MAX_PATH
#endif /* __WIN32_DW */

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "dwc_lib.h"
#include "dwc_err.h"
#include "dwc_macro.h"
#include "dwc_type.h"
#include "dwc_thread.h"

#if defined(__WIN32_DW)
#define Snprintf            _snprintf
#endif /* __WIN32_DW */
#if defined(__UNIX_DW)
#define Snprintf            snprintf
#endif /* __UNIX_DW */

#define KEY_LEN                 32
#define LOGLINE_LEN             4096

#define DWCLIENT_NOERROR        0     /* 0x00000000 */
#define DWCLIENT_EXPAND         1<<31 /* 0x80000000 */
#define DWCLIENT_DEMO_USAGE     1<<30 /* 0x40000000 */
#define DWCLIENT_NO_MEMORY      1<<29 /* 0x20000000 */
#define DWCLIENT_LIB_ERR        1<<28 /* 0x10000000 */

#define DWANS_OK                0
#define DWANS_VIRUS_FOUND       1
#define DWANS_MAILBOMB          2
#define DWANS_SKIPOBJECT        3
#define DWANS_ERROR             4
#define DWANS_TEMPFAIL          5
#define DWANS_SPAM              6

typedef struct {
    char            *filename;
    unsigned long    id;
} DWL_scan_request_t;

static void     scanner( void* ptr );
static int      is_key_req( char* req_str );
static int      is_file_req( char* req_str );
static void     dispatch_unhandled( char* req_str );
static void     put_answer( unsigned long rid, int op, int code, char* msg );
static int      write_string( char* msg );
static void     replace_spec_chars( char* msg, char* nmsg );
static int      parseCmdLine( int argc, char ** argv );
static int      usage( void );

static char szRequest[ PATH_MAX + 16 ] = { 0 };
static char szDaemonIP[ PATH_MAX + 2 ] = { 0 };
static char szSocketFile[ PATH_MAX + 2 ] = { 0 };
static char szPidFile[ PATH_MAX + 2 ] = { 0 };
static char szKeyFormat[ 64 ] = { 0 };
static char szFileFormat[ 64 ] = { 0 };
static int  bCache = 0;
static int  bQuiet = 0;
static int  bDebug = 0;
#if defined(__UNIX_DW)
static int  bLocalFile = 0;
#endif /* __UNIX_DW */
static int  scanOptions = DRWEBD_MAIL_MESSAGE;
static int  Port = 0;
static int  Timeout = 0;
static int  ActionOnMailbomb = DWANS_MAILBOMB;
static int  ActionOnSkipObject = DWANS_OK;
static int  ActionOnScanError = DWANS_ERROR;
static int  ActionOnError = DWANS_ERROR;

/* ------------------------------------------------------------------------- */
char get_action( int action )
{
    char    ret = 'U';

    switch( action )
    {
     case DWANS_OK:
        ret = 'A';
        break;
     case DWANS_ERROR:
     case DWANS_MAILBOMB:
     case DWANS_SKIPOBJECT:
     case DWANS_VIRUS_FOUND:
        ret = 'R';
        break;
     case DWANS_TEMPFAIL:
        ret = 'T';
        break;
    }
    return ret;
}

/* ------------------------------------------------------------------------- */
int main( int argc, char *argv[] )
{
    int retcode = DWCLIENT_NOERROR;
    int libcode = DWL_NOERROR;

    if( !parseCmdLine( argc, argv ) )
    {
        return usage();
    }
#if defined(__UNIX_DW)
    outputLib( bQuiet, bDebug, 1 );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    outputLib( 1, 0, 0 );
#endif /* __UNIX_DW */
    dprintLib("DrWeb-CGP: output initialized -> trying to start.\n");
    if( Snprintf( szKeyFormat, sizeof(szKeyFormat), "%%d KEY %%%ds %%%ds", KEY_LEN, KEY_LEN ) < 0 )
    {
        dprintLib("DrWeb-CGP: can not create KEY request format.\n");
        return 1;
    }
    if( Snprintf( szFileFormat, sizeof(szFileFormat), "%%d FILE %%%ds", PATH_MAX ) < 0 )
    {
        dprintLib("DrWeb-CGP: can not create FILE request format.\n");
        return 1;
    }
    if( (szDaemonIP[0] != NULLCHAR) || (Port > 0) )
    {
        libcode = startupLib( DWL_TYPE_TCP_SOCKET, Timeout, szDaemonIP, Port, bCache );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("DrWeb-CGP: can not initialize library with TCP/IP transport.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
#if defined(__UNIX_DW)
    else if( szSocketFile[0] != NULLCHAR )
    {
        libcode = startupLib( DWL_TYPE_UNIX_SOCKET, Timeout, szSocketFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("DrWeb-CGP: can not initialize library with UNIX sockets transport.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    else if( szPidFile[0] != NULLCHAR )
    {
        libcode = startupLib( DWL_TYPE_READPIDFILE, Timeout, szPidFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("DrWeb-CGP: can not initialize library (pidfile reading error).\n");
            return DWCLIENT_LIB_ERR;
        }
    }
#endif /* __UNIX_DW */
    else
    {
        dprintLib("DrWeb-CGP: daemon address is not defined, defaults will be used.\n");
        libcode = startupLib( DWL_TYPE_DEFAULT, Timeout );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("DrWeb-CGP: can not initialize library.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    dprintLib("DrWeb-CGP: Formats: KEY = %s, FILE = %s\n", szKeyFormat, szFileFormat );
    dprintLib("DrWeb-CGP: Actions: skip=%c, bomb=%c, error=%c, fatal=%c.\n", get_action(ActionOnSkipObject), get_action(ActionOnMailbomb), get_action(ActionOnScanError), get_action(ActionOnError) );
    wprintLib("DrWeb-CGP: started!\n");
    while( fgets( szRequest, sizeof(szRequest), stdin ) != NULL )
    {
        dprintLib("DrWeb-CGP: read - %s", szRequest);
        if( is_file_req( szRequest ) )
        {
            dprintLib("DrWeb-CGP: new scan request scan.\n");
        }
        else if( is_key_req( szRequest ) )
        {
            dprintLib("DrWeb-CGP: key request.\n");
        }
        else
        {
            dprintLib("DrWeb-CGP: unhandled request.\n");
            dispatch_unhandled( szRequest );
        }
    }
    wprintLib("DrWeb-CGP: shutdown in progress ...");
    cleanupLib();
    wprintLib("DrWeb-CGP: shutdown OK!");
    return retcode;
}

/* ------------------------------------------------------------------------- */
static void dispatch_unhandled( char* req_str )
{
    int id;

    if( sscanf( req_str, "%d", &id ) == 1 )
    {
        put_answer( id, DWANS_ERROR, errno, "DrWeb Client Library: unknown CommuniGate Pro request" );
    }
    return;
}

/* ------------------------------------------------------------------------- */
static int is_key_req( char* req_str )
{
    char key1[KEY_LEN] = { 0 };
    char key2[KEY_LEN] = { 0 };
    int  req;

    if( sscanf( req_str, szKeyFormat, &req, key1, key2 ) != 3 )
    {
        return 0;
    }
    /* asv: I havent ideas what do with keys - dummy answer 'OK' */
    dprintLib("DrWeb-CGP: passed req -> { id=%d key=<%s %s> }.\n", req, key1, key2 );
    put_answer( req, DWANS_OK, 0, NULL );
    return 1;
}

/* ------------------------------------------------------------------------- */
static int is_file_req( char* req_str )
{
    DWL_scan_request_t  *req = NULL;
    DWC_error_code_t     err;

    req = DWC_Malloc( DWL_scan_request_t*, sizeof(DWL_scan_request_t) );
    if( !req )
    {
        wprintLib("DrWeb-CGP: no memory.\n");
        return 0;
    }
    req->filename = DWC_Malloc( char*, PATH_MAX+1 );
    if( !req->filename )
    {
        wprintLib("DrWeb-CGP: no memory.\n");
        DWC_Free( req );
        return 0;
    }
    if( sscanf( req_str, szFileFormat, &req->id, req->filename ) != 2 )
    {
        dprintLib("DrWeb-CGP: is not FILE command.\n");
        DWC_Free( req->filename );
        DWC_Free( req );
        return 0;
    }
    dprintLib( "DrWeb-CGP: passed req (%p) -> { id=%lu file=%s }\n", req, req->id, req->filename );
    err = dwc_thread_create( scanner, req );
    if( err != DWC_NOERROR )
    {
        wprintLib("DrWeb-CGP: can not create scanning thread.\n");
        put_answer( req->id, ActionOnError, 0, NULL );
        DWC_Free( req->filename );
        DWC_Free( req );
    }
    return 1;
}

/* ------------------------------------------------------------------------- */
static void scanner( void* ptr )
{
    int                 retcode = DERR_NOERROR;
    int                 libcode = DWL_NOERROR;
    char               *plog = NULL;
    size_t             *plen = NULL;
    DWL_scan_request_t *req = (DWL_scan_request_t*)ptr;
    size_t              loglen;
#if defined(__UNIX_DW)
    char                *buf;
#endif /* __UNIX_DW */

    if( !req || !req->filename )
    {
        wprintLib( "ScanThread: null pointer passed.\n" );
        /* asv: we sure that req->filename doesnt exist
           and null ptr freeing is safe */
        DWC_Free( req );
        return;
    }
    dprintLib( "ScanThread: called for req(%p) -> { id=%lu name=%s }\n", req, req->id, req->filename );

    if( scanOptions & DRWEBD_RETURN_VIRUSES )
    {
        loglen = LOGLINE_LEN;
        plog = DWC_Malloc( char*, loglen );
        if( plog )
        {
            plen = &loglen;
        }
        else
        {
            plen = NULL;
        }
    }
    else
    {
        plog = NULL;
        plen = NULL;
    }

#if defined(__UNIX_DW)
    if( bLocalFile )
    {
        buf = DWC_Malloc( char*, PATH_MAX + 1 );
        if( !buf )
        {
            wprintLib("ScanThread: no memory.\n");
            retcode = DWCLIENT_NO_MEMORY;
            goto error_exit;
        }
        if( realpath( req->filename, buf ) == NULL )
        {
            wprintLib("ScanThread: can not get real path to %s.\n", req->filename);
            DWC_Free( buf );
            retcode = DWCLIENT_EXPAND;
            goto error_exit;
        }
        dprintLib("ScanThread: %s expanded to %s.\n", req->filename, buf);
        DWC_Free( req->filename );
        req->filename = buf;
        libcode = scanFile( req->filename, strlen(req->filename), DWL_FT_DISKFILE, scanOptions, &retcode, plog, plen );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_LIB_ERR;
        }
    }
    else
#endif /* __UNIX_DW */
    {
        libcode = scanFile( req->filename, strlen(req->filename), DWL_FT_NORMALFILE, scanOptions, &retcode, plog, plen );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_LIB_ERR;
        }
    }
  error_exit:
    if( (retcode & ~DERR_MASK ) != 0 ) /* have local errors */
    {
        if( retcode & DWCLIENT_LIB_ERR )
        {
            dprintLib("ScanThread: dwclient library error %d (%s)\n", libcode, getLibError( libcode ) );
        }
        else
        {
            dprintLib("ScanThread: dwclient error %d\n", retcode );
        }
        put_answer( req->id, ActionOnScanError, 0, NULL );
    }
    else
    {
        if( (retcode != DERR_NOERROR) && (retcode != DERR_EVAL_KEY) ) /* scan not ok - errors or viruses */
        {
            wprintLib("ScanThread: daemon return code %#x (%s)\n", retcode, getDaemonCode( retcode ) );
            if( retcode & DERR_VIRUS_MASK )
            {
                put_answer( req->id, DWANS_VIRUS_FOUND, 0, plog );
            }
            else if( retcode & DERR_SPAM )
            {
                put_answer( req->id, DWANS_SPAM, 0, NULL );
            }
            else if( retcode & DERR_MAIL_BOMB )
            {
                put_answer( req->id, ActionOnMailbomb, 0, NULL );
            }
            else if( retcode & DERR_SKIP_OBJECT )
            {
                put_answer( req->id, ActionOnSkipObject, 0, NULL );
            }
            else
            {
#if defined(__UNIX_DW)
                if( bLocalFile && (retcode == (DERR_WRITE_ERR | DERR_EVAL_KEY)) )
                {
                    dprintLib("for EVALUATION key and <localscan> mode CommuniGate QUEUE dir should writable for daemon\n");
                }
#endif /* __UNIX_DW */
                if( ActionOnScanError == DWANS_OK )
                    put_answer( req->id, ActionOnScanError, 0, NULL );
                else
                    put_answer( req->id, ActionOnScanError, retcode, getDaemonCode( retcode ) );
            }
        }
        else
        {
            dprintLib("ScanThread: daemon return (%#x) is okey.\n", retcode);
            put_answer( req->id, DWANS_OK, 0, NULL );
        }
    }
    DWC_Free( plog );
    DWC_Free( req->filename );
    DWC_Free( req );
    return;
}

/* ------------------------------------------------------------------------- */
static void put_answer( unsigned long rid, int op, int code, char* msg )
{
    char   buf[255] = { 0 };
    char  *nmsg;

    if( msg )
    {
        nmsg = DWC_Malloc( char*, 2*(strlen(msg)+1) );
        if( nmsg )
        {
            replace_spec_chars( msg, nmsg );
        }
    }
    else
    {
        nmsg = NULL;
    }
    if( op == DWANS_OK )
    {
        Snprintf( buf, sizeof(buf), "%lu OK\n", rid );
    }
    else if( op == DWANS_VIRUS_FOUND )
    {
        if( nmsg )
            Snprintf( buf, sizeof(buf), "%lu ERROR \"A virus(es) are found in your message.\\eDiagnostic:\\e%s\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid, nmsg );
        else
            Snprintf( buf, sizeof(buf), "%lu ERROR \"A virus(es) are found in your message.\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid );
    }
    else if( op == DWANS_SPAM )
    {
        if( nmsg )
            Snprintf( buf, sizeof(buf), "%lu ERROR \"You sent a message that looks like a spam.\\eDiagnostic:\\e%s\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid, nmsg );
        else
            Snprintf( buf, sizeof(buf), "%lu ERROR \"You sent a message that looks like a spam.\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid );
    }
    else if( op == DWANS_MAILBOMB )
    {
        if( nmsg )
            Snprintf( buf, sizeof(buf), "%lu ERROR \"Possible a mailbomb are contained in your message.\\eDiagnostic:\\e%s\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid, nmsg );
        else
            Snprintf( buf, sizeof(buf), "%lu ERROR \"Possible a mailbomb are contained in your message.\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid );
    }
    else if( op == DWANS_SKIPOBJECT )
    {
        if( nmsg )
            Snprintf( buf, sizeof(buf), "%lu ERROR \"Your message contains object(s) that cannot be checked by antivirus.\\eDiagnostic:\\e%s\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid, nmsg );
        else
            Snprintf( buf, sizeof(buf), "%lu ERROR \"Your message contains object(s) that cannot be checked by antivirus.\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid );
    }
    else if( op == DWANS_ERROR )
    {
        if( nmsg )
            Snprintf( buf, sizeof(buf), "%lu ERROR \"Your message has not been delivery because cannot be checked by antivirus.\\eDiagnostic:\\eError %d : %s\\ePlease contact to postmaster.\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid, code, nmsg );
        else
            Snprintf( buf, sizeof(buf), "%lu ERROR \"Your message has not been delivery because cannot be checked by antivirus.\\e Please contact to postmaster.\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid );
    }
    else if( op == DWANS_TEMPFAIL )
    {
        if( nmsg )
            Snprintf( buf, sizeof(buf), "%lu REJECTED \"Your message has been temporary failed to delivery.\\ePlease contact postmaster if it is not first notification\\eDiagnostic:\\e%s\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid, nmsg );
        else
            Snprintf( buf, sizeof(buf), "%lu REJECTED \"Your message has been temporary failed to delivery.\\ePlease contact postmaster if it is not first notification\\e \\tAntivirus service provided by DrWeb Daemon (www.sald.com).\"\n", rid );
    }
    else
    {
        wprintLib( "PutAnswer: unknown opcode.\n" );
        DWC_Free( nmsg );
        return;
    }
    buf[sizeof(buf)-1] = '\000';
    write_string( buf );
    DWC_Free( nmsg );
    return;
}

/* ------------------------------------------------------------------------- */
static void replace_spec_chars( char* msg, char* nmsg )
{
     do
     {
         switch( *msg )
         {
             case '\r':
             case '\n':
                 /* replace \r or \n -> "\\e" (CGP End-of-Line) */
                 *nmsg = '\\';
                 ++nmsg;
                 *nmsg = 'e';
                 break;
             case '\t':
                 /* replace \t -> "\\t" (CGP Tab) */
                 *nmsg = '\\';
                 ++nmsg;
                 *nmsg = 't';
                 break;
             case '"':
                 /* replace \ -> \" (CGP backslash) */
                 *nmsg = '\\';
                 ++nmsg;
                 *nmsg = '"';
                 break;
             default:
                 /* just copy other chars */
                 *nmsg = *msg;
                 break;
         }
         ++msg;
         ++nmsg;
    } while( *msg );
    *nmsg = '\0';
}
/* ------------------------------------------------------------------------- */
static int write_string( char* msg )
{
    int    written, overall, msg_len;

    for( overall = 0, msg_len = strlen(msg); overall < msg_len; overall += written )
    {
        written = write( fileno(stdout), &msg[overall], msg_len - overall );
        if( written < 0 )
        {
            wprintLib( "WriteString: error on writing to stdout.\n" );
            return -1;
        }
        else if( written != msg_len )
        {
            wprintLib( "WriteString: possible problems with wrining to stdout.\n" );
        }
    }
    return overall;
}

/* ------------------------------------------------------------------------- */
static int parseCmdLine( int argc, char ** argv )
{
#if defined(__UNIX_DW) && !defined(EMX_DW)
    int  intVal, Cont;

    Cont = 1;
    while( Cont )
    {
        switch( getopt( argc, argv, "hlqrvwd:u:n:p:t:r:b:y:j:k:") )
        {
            case 'd':
                strncpy( szPidFile, optarg, sizeof(szPidFile)-1 );
                break;
            case 'u':
                strncpy( szSocketFile, optarg, sizeof(szSocketFile)-1 );
                break;
            case 'n':
                strncpy( szDaemonIP, optarg, sizeof(szDaemonIP)-1 );
                break;
            case 'p':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("Error reading port argument (-p option).\n");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Port = intVal;
                }
                break;
            case 't':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("Error reading timeout argument (-t timeout).\n");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Timeout = intVal;
                }
                break;
            case 'h':
                scanOptions |= DRWEBD_HEURISTIC_ON;
                break;
            case 'm':
                scanOptions |= DRWEBD_SPAM_FILTER;
                break;
            case 'r':
                scanOptions |= DRWEBD_RETURN_VIRUSES;
                break;
            case 'b':
                switch( optarg[0] )
                {
                    case 'a':
                        ActionOnError = DWANS_OK;
                        break;
                    case 'r':
                        ActionOnError = DWANS_ERROR;
                        break;
                    case 't':
                        ActionOnError = DWANS_TEMPFAIL;
                        break;
                    default:
                        printf("Invalid argument for option -b");
                        return 0;
                }
                break;
            case 'y':
                switch( optarg[0] )
                {
                    case 'a':
                        ActionOnScanError = DWANS_OK;
                        break;
                    case 'r':
                        ActionOnScanError = DWANS_ERROR;
                        break;
                    case 't':
                        ActionOnScanError = DWANS_TEMPFAIL;
                        break;
                    default:
                        printf("Invalid argument for option -y");
                        return 0;
                }
                break;
            case 'j':
                switch( optarg[0] )
                {
                    case 'a':
                        ActionOnMailbomb = DWANS_OK;
                        break;
                    case 'r':
                        ActionOnMailbomb = DWANS_MAILBOMB;
                        break;
                    default:
                        printf("Invalid argument for option -j");
                        return 0;
                }
                break;
            case 'k':
                switch( optarg[0] )
                {
                    case 'a':
                        ActionOnSkipObject = DWANS_OK;
                        break;
                    case 'r':
                        ActionOnSkipObject = DWANS_SKIPOBJECT;
                        break;
                    default:
                        printf("Invalid argument for option -k");
                        return 0;
                }
                break;
            case 'l':
                bLocalFile = 1;
                break;
            case 'v':
                bCache = 1;
                break;
            case 'q':
                bQuiet = 1;
                break;
            case 'w':
                bDebug = 1;
                break;
            case -1 :
                Cont = 0;
                break;
            default :
                return 0;
        }
    }
#else /* !UNIX || EMX */
    int i;

    for( i = 1; i < argc; i++ )
    {
        if( !strncmp( *(argv + i), "-n", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            strncpy( szDaemonIP, *( argv + i ) + 2, PATH_MAX );
        }
        else if( !strncmp( *(argv + i), "-p", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            Port = atoi( *( argv + i ) + 2 );
        }
        else if( !strncmp( *(argv + i), "-t", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            Timeout = atoi( *( argv + i ) + 2 );
        }
        else if( !strncmp( *(argv + i), "-h", 2 ) )
        {
            scanOptions |= DRWEBD_HEURISTIC_ON;
        }
        else if( !strncmp( *(argv + i), "-m", 2 ) )
        {
            scanOptions |= DRWEBD_SPAM_FILTER;
        }
        else if( !strncmp( *(argv + i), "-r", 2 ) )
        {
            scanOptions |= DRWEBD_RETURN_VIRUSES;
        }
        else if( !strncmp( *(argv + i), "-v", 2 ) )
        {
            bCache = 1;
        }
        else if( !strncmp( *(argv + i), "-ba", 2 ) )
        {
            ActionOnError = DWANS_OK;
        }
        else if( !strncmp( *(argv + i), "-br", 2 ) )
        {
            ActionOnError = DWANS_ERROR;
        }
        else if( !strncmp( *(argv + i), "-bt", 2 ) )
        {
            ActionOnError = DWANS_TEMPFAIL;
        }
        else if( !strncmp( *(argv + i), "-ya", 2 ) )
        {
            ActionOnScanError = DWANS_OK;
        }
        else if( !strncmp( *(argv + i), "-yr", 2 ) )
        {
            ActionOnScanError = DWANS_ERROR;
        }
        else if( !strncmp( *(argv + i), "-yt", 2 ) )
        {
            ActionOnScanError = DWANS_TEMPFAIL;
        }
        else if( !strncmp( *(argv + i), "-ja", 2 ) )
        {
            ActionOnMailbomb = DWANS_OK;
        }
        else if( !strncmp( *(argv + i), "-jr", 2 ) )
        {
            ActionOnMailbomb = DWANS_MAILBOMB;
        }
        else if( !strncmp( *(argv + i), "-ka", 2 ) )
        {
            ActionOnSkipObject = DWANS_OK;
        }
        else if( !strncmp( *(argv + i), "-kr", 2 ) )
        {
            ActionOnSkipObject = DWANS_SKIPOBJECT;
        }
        else if( !strncmp( *(argv + i), "-q", 2 ) )
        {
            bQuiet = 1;
        }
        else if( !strncmp( *(argv + i), "-w", 2 ) )
        {
            bDebug = 1;
        }
        else
        {
            return 0;
        }
    }
#endif /* __UNIX_DW */
    return 1;
}

/* ------------------------------------------------------------------------- */
static int usage( void )
{
    printf( "\nUsage: drweb-cgp [options]"
            "\n"
#if defined(__UNIX_DW) && !defined(EMX_DW)
            "\n -dPIDFILE_NAME   Dr.Web daemon PID-file (/var/run/drwebd.pid is default)"
            "\n -uSOCKET_FILE    Dr.Web daemon UNIX-socket file"
#endif /* __UNIX_DW && !EMX */
            "\n -nINETADDRESS    Dr.Web daemon Internet address (IP or domain name)"
            "\n -pPORT           Dr.Web daemon TCP port"
            "\n -tTIMEOUT        timeout (millisecods) for communcations operations"
            "\n -h               use heuristic analyser"
            "\n -m               use spam analyser"
            "\n -r               put virus names into notification"
            "\n -bACTION         filter action on fatal errors in filter"
            "\n                  (mostly memory allocation or configuration problems)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject (default) | a - accept | t - tempfail }"
            "\n -yACTION         filter action on errors during scanning"
            "\n                  (timeouts or memory allocation problems for example)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject (default) | a - accept | t - tempfail }"
            "\n -jACTION         filter action on possible mailbomb in message"
            "\n                  (mostly archive with compression > MaxCompressionRation)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject (default) | a - accept }"
            "\n -kACTION         filter action on skipped object in message"
            "\n                  (mostly archives: passworded or part of multivolume)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject | a - accept (default) }"
            "\n -v               use cached ip-address for daemon (resolved at start)"
#if defined(__UNIX_DW) && !defined(EMX_DW)
            "\n -l               allow daemon to scan messages as local files"
#endif /* __UNIX_DW && !EMX */
#if defined(__UNIX_DW)
            "\n -q               quiet mode"
            "\n -w               debug mode"
#endif /* __UNIX_DW */
            "\n" );
    return DWCLIENT_DEMO_USAGE;
}
