/*--------------------------------------------------------------------------
 Client for Dr.Web daemon
                                       ID Antivirus Lab http://www.drweb.ru
                                        http://www.drweb.ru/linux-bsd/forum
                                              http://www.drweb.ru/linux-bsd
--------------------------------------------------------------------------*/

#ifndef lint
static const char rcsid[] = "$Id: drwebdc.c,v 1.30 2001/10/16 15:28:21 cvser Exp $";
#endif

#if defined(__WIN32_DW)
#include <io.h>
#endif /* __WIN32_DW */
#if defined(__UNIX_DW)
#include <unistd.h>
#include <sys/param.h>
#endif /* __UNIX_DW */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include "dwc_lib.h"
#include "dwc_macro.h"

#if defined(__WIN32_DW)
#define PATH_MAX    _MAX_PATH
#endif /* __WIN32_DW */

/* Bits in return code (8-bit) */
#define DWCLIENT_NOERROR        (0<<0)   /* 0x00 -   0 */
#define DWCLIENT_VIRUSFOUND     (1<<0)   /* 0x01 -   1 */
#define DWCLIENT_SKIPOBJECT     (1<<1)   /* 0x02 -   2 */
#define DWCLIENT_MAILBOMB       (1<<2)   /* 0x04 -   4 */
#define DWCLIENT_DAEMONERROR    (1<<3)   /* 0x08 -   8 */
#define DWCLIENT_CLIENTERROR    (1<<4)   /* 0x10 -  16 */
#define DWCLIENT_EVALKEY        (1<<5)   /* 0x20 -  32 */
#define DWCLIENT_SPAM           (1<<6)   /* 0x40 -  64 */
#define DWCLIENT_USAGE          (1<<7)   /* 0x80 - 128 */

static char szFile[ PATH_MAX + 2 ] = { 0 };
static char szDaemonIP[ PATH_MAX + 2 ] = { 0 };
static char szSocketFile[ PATH_MAX + 2 ] = { 0 };
static char szPidFile[ PATH_MAX + 2 ] = { 0 };
static int  scanOptions = 0;
static int  Port = 0;
static int  Timeout = 0;
static int  bShowID = 0;
static int  bShowVersion = 0;
static int  bShowBases = 0;
static int  bScanFile = 0;
static int  bQuiet = 0;
static int  bDebug = 0;
#if defined(__UNIX_DW)
static int  bPartFile = 0;
static int  bLocalFile = 0;
static size_t Offset = 0;
static size_t Length = 0;
#endif /* __UNIX_DW */

static int  parseCmdLine( int argc, char ** argv );
static int  usage( void );

#if defined(__WIN32_DW)
#define read_file( fd, buf, len )   _read( (fd), (buf), (len) )
#else /* __WIN32_DW */
#define read_file( fd, buf, len )   read( (fd), (buf), (len) )
#endif /* __WIN32_DW */

/* ---------------------------------------------------------------------- */
int main( int argc, char *argv[] )
{
    int retcode = DWCLIENT_NOERROR;
    int libcode = DWL_NOERROR;
    int newcode;
    char *pFileBuffer = NULL;
    char *pTmp = NULL;

    if( !parseCmdLine( argc, argv ) )
    {
        return usage();
    }
    outputLib( bQuiet, bDebug, 0 );
    if( (szDaemonIP[0] != NULLCHAR) || (Port > 0) )
    {
        libcode = startupLib( DWL_TYPE_TCP_SOCKET, Timeout, szDaemonIP, Port, 0 );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can not initialize library with TCP/IP transport.\n");
            return DWCLIENT_CLIENTERROR;
        }
    }
#if defined(__UNIX_DW)
    else if( szSocketFile[0] != NULLCHAR )
    {
        libcode = startupLib( DWL_TYPE_UNIX_SOCKET, Timeout, szSocketFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can not initialize library with UNIX sockets transport.\n");
            return DWCLIENT_CLIENTERROR;
        }
    }
    else if( szPidFile[0] != NULLCHAR )
    {
        libcode = startupLib( DWL_TYPE_READPIDFILE, Timeout, szPidFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can not initialize library (pidfile reading error).\n");
            return DWCLIENT_CLIENTERROR;
        }
    }
#endif /* __UNIX_DW */
    else
    {
        dprintLib("Daemon address is not defined, default will be used.\n");
        libcode = startupLib( DWL_TYPE_DEFAULT, Timeout );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Library initialization error.\n");
            return DWCLIENT_CLIENTERROR;
        }
    }
    if( bShowVersion )
    {
        libcode = showVersion( NULL, NULL );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_CLIENTERROR;
            goto error_exit;
        }
    }
    if( bShowID )
    {
        libcode = showID( NULL, NULL );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_CLIENTERROR;
            goto error_exit;
        }
    }
    if( bShowBases )
    {
        libcode = showBases( NULL, NULL );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_CLIENTERROR;
            goto error_exit;
        }
    }
    if( bScanFile )
    {
        if( szFile[0] != NULLCHAR )
        {
#if defined(__UNIX_DW)
            if( bLocalFile || bPartFile )
            {
                char buf[ PATH_MAX + 2 ];
                if( realpath( szFile, buf ) == NULL )
                {
                    wprintLib("Can not expand %s.\n", szFile);
                    retcode = DWCLIENT_CLIENTERROR;
                    goto error_exit;
                }
                dprintLib( "Filename %s expanded to %s\n", szFile, buf );
                strncpy( szFile, buf, PATH_MAX + 1 );
           }
           if( bLocalFile )
           {
                libcode = scanFile( szFile, strlen(szFile), DWL_FT_DISKFILE, scanOptions, &retcode, NULL, NULL );
                if( libcode != DWL_NOERROR )
                {
                    retcode = DWCLIENT_CLIENTERROR;
                }
            }
            else if( bPartFile )
            {
                libcode = scanFilePart( szFile, strlen(szFile), Offset, Length, scanOptions, &retcode, NULL, NULL );
                if( libcode != DWL_NOERROR )
                {
                    retcode = DWCLIENT_CLIENTERROR;
                }
            }
            else
#endif /* __UNIX_DW */
            {
                libcode = scanFile( szFile, strlen(szFile), DWL_FT_NORMALFILE, scanOptions, &retcode, NULL, NULL );
                if( libcode != DWL_NOERROR )
                {
                    retcode = DWCLIENT_CLIENTERROR;
                }
            }
        }
        else
        {
            char buf[ 0x4000 ];
            int fileSize = 0;
            int readed = read_file( fileno(stdin), buf, sizeof(buf) );
            while( readed > 0 )
            {
                pTmp = realloc( pFileBuffer, fileSize + readed );
                if( !pTmp )
                {
                    wprintLib("No memory for file reading.\n");
                    free( pFileBuffer );
                    retcode = DWCLIENT_CLIENTERROR;
                    goto error_exit;
                }
                memcpy( pTmp + fileSize, buf, readed );
                pFileBuffer = pTmp;
                fileSize += readed;
                readed = read_file( fileno(stdin), buf, sizeof(buf) );
            }
            libcode = scanBuffer( pFileBuffer, fileSize, scanOptions, &retcode, NULL, NULL );
            if( libcode != DWL_NOERROR )
            {
                retcode = DWCLIENT_CLIENTERROR;
            }
        }
    }

  error_exit:
    cleanupLib();
    if( pFileBuffer )
    {
        free( pFileBuffer );
        pFileBuffer = NULL;
    }
    if( retcode != DWCLIENT_CLIENTERROR )
    {
        /* repack daemon error codes to 8-bit value */
        dprintLib( "Results: daemon return code %#x (%s)\n", retcode, getDaemonCode( retcode ) );
        newcode = DWCLIENT_NOERROR;
        if( retcode & DERR_VIRUS_MASK )
        {
            newcode |= DWCLIENT_VIRUSFOUND;
        }
        if( retcode & DERR_SKIP_OBJECT )
        {
            newcode |= DWCLIENT_SKIPOBJECT;
        }
        if( retcode & DERR_MAIL_BOMB )
        {
            newcode |= DWCLIENT_MAILBOMB;
        }
        if( retcode & DERR_DAEMON_ERROR )
        {
            newcode |= DWCLIENT_DAEMONERROR;
        }
        if( retcode & DERR_EVAL_KEY )
        {
            newcode |= DWCLIENT_EVALKEY;
        }
        if( retcode & DERR_SPAM )
        {
            newcode |= DWCLIENT_SPAM;
        }
        retcode = newcode;
    }
    else
    {
        dprintLib( "Results: dwclient error %d (%s)\n", libcode, getLibError( libcode ) );
    }
    dprintLib( "Final code %#x.\n", retcode );
    return retcode;
}

/* ---------------------------------------------------------------------- */
static int parseCmdLine( int argc, char ** argv )
{
#if defined(__UNIX_DW) && !defined(EMX_DW)

#define DWC_PCO_NAME    (1<<0)
#define DWC_PCO_OFF     (1<<1)
#define DWC_PCO_LEN     (1<<2)
#define DWC_PCO_FULL    (DWC_PCO_NAME | DWC_PCO_OFF | DWC_PCO_LEN)
    int  PartCmdOpt = 0, intVal, Cont;

    Cont = 1;
    while( Cont )
    {
        switch( getopt( argc, argv, "xmihqws:r:t:n:p:d:u:c:o:l:f:") )
        {
            case 'f':
                if( bScanFile )
                {
                    printf("\nYou can not use -f and -i option together.");
                    return 0;
                }
                if( optarg )
                {
                    strncpy( szFile, optarg, sizeof(szFile)-1 );
                }
                bScanFile = 1;
                break;
            case 'i':
                if( bScanFile )
                {
                    printf("\nYou can not use -f and -i option together.");
                    return 0;
                }
                bScanFile = 1;
                break;
            case 'd':
                strncpy( szPidFile, optarg, sizeof(szPidFile)-1 );
                break;
            case 'u':
                strncpy( szSocketFile, optarg, sizeof(szSocketFile)-1 );
                break;
            case 'c':
                strncpy( szFile, optarg, sizeof(szFile)-1 );
                PartCmdOpt |= DWC_PCO_NAME;
                bScanFile = 1;
                bPartFile = 1;
                break;
            case 'o':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("\nError reading offset argument (-o option).");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Offset = intVal;
                    PartCmdOpt |= DWC_PCO_OFF;
                }
                break;
            case 'l':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("\nError reading length argument (-l option).");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Length = intVal;
                    PartCmdOpt |= DWC_PCO_LEN;
                }
                break;
            case 'n':
                strncpy( szDaemonIP, optarg, sizeof(szDaemonIP)-1 );
                break;
            case 'p':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("\nError reading port argument (-p option).");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Port = intVal;
                }
                break;
            case 't':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("\nError reading timeout argument (-t timeout).");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Timeout = intVal;
                }
                break;
            case 'h':
                scanOptions |= DRWEBD_HEURISTIC_ON;
                break;
            case 'm':
                scanOptions |= DRWEBD_SPAM_FILTER;
                break;
            case 'x':
                bLocalFile = 1;
                break;
            case 's':
                switch( optarg[0] )
                {
                    case 'v':
                        bShowVersion = 1;
                        break;
                    case 'i':
                        bShowID = 1;
                        break;
                    case 'b':
                        bShowBases = 1;
                        break;
                    default:
                        printf("\nInvalid option show (-s?)");
                        return 0;
                }
                break;
            case 'r':
                switch( optarg[0] )
                {
                    case 'v':
                        scanOptions |= DRWEBD_RETURN_VIRUSES;
                        break;
                    case 'c':
                        scanOptions |= DRWEBD_RETURN_CODES;
                        break;
                    case 'r':
                        scanOptions |= DRWEBD_RETURN_REPORT;
                        break;
                    default:
                        printf("\nInvalid option return (-r?)");
                        return 0;
                }
                break;
            case 'q':
                bQuiet = 1;
                break;
            case 'w':
                bDebug = 1;
                break;
            case -1 :
                if( bPartFile && ((PartCmdOpt & DWC_PCO_FULL) != DWC_PCO_FULL) )
                {
                    printf( "\nFor scan only part file you SHOULD define -c -o -l options." );
                    return 0;
                }
                Cont = 0;
                break;
            default :
                return 0;
        }
    }
#else /* !UNIX || EMX */
    int i;

    for( i = 1; i < argc; i++ )
    {
        if( !strncmp( *(argv + i), "-f", 2 ) )
        {
            if( strlen( *(argv + i) ) > 2 )
            {
                strncpy( szFile, *( argv + i ) + 2, PATH_MAX );
            }
            bScanFile = 1;
        }
        else if( !strncmp( *(argv + i), "-i", 2 ) )
        {
            szFile[0] = NULLCHAR;
            bScanFile = 1;
        }
        else if( !strncmp( *(argv + i), "-n", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            strncpy( szDaemonIP, *( argv + i ) + 2, PATH_MAX );
        }
        else if( !strncmp( *(argv + i), "-p", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            Port = atoi( *( argv + i ) + 2 );
        }
        else if( !strncmp( *(argv + i), "-t", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            Timeout = atoi( *( argv + i ) + 2 );
        }
        else if( !strncmp( *(argv + i), "-h", 2 ) )
        {
            scanOptions |= DRWEBD_HEURISTIC_ON;
        }
        else if( !strncmp( *(argv + i), "-m", 2 ) )
        {
            scanOptions |= DRWEBD_SPAM_FILTER;
        }
        else if( !strncmp( *(argv + i), "-rv", 3 ) )
        {
            scanOptions |= DRWEBD_RETURN_VIRUSES;
        }
        else if( !strncmp( *(argv + i), "-rr", 3 ) )
        {
            scanOptions |= DRWEBD_RETURN_REPORT;
        }
        else if( !strncmp( *(argv + i), "-rc", 3 ) )
        {
            scanOptions |= DRWEBD_RETURN_CODES;
        }
        else if( !strncmp( *(argv + i), "-q", 2 ) )
        {
            bQuiet = 1;
        }
        else if( !strncmp( *(argv + i), "-w", 2 ) )
        {
            bDebug = 1;
        }
        else if( !strncmp( *(argv + i), "-sb", 3 ) )
        {
            bShowBases = 1;
        }
        else if( !strncmp( *(argv + i), "-si", 3 ) )
        {
            bShowID = 1;
        }
        else if( !strncmp( *(argv + i), "-sv", 3 ) )
        {
            bShowVersion = 1;
        }
#ifdef EMX_DW
        else if( !strncmp( *(argv + i), "-x", 2 ) )
        {
            bLocalFile = 1;
        }
#endif /* EMX */
        else
        {
            return 0;
        }
    }
#endif /* __UNIX_DW */
    return argc > 1;
}

/* ---------------------------------------------------------------------- */
static int usage( void )
{
    printf( "\nUsage: drwebdc [options]"
            "\n"
            "\n -fFILE_NAME      file to scan"
            "\n -i               read data from stdin"
#if defined(__UNIX_DW) && !defined(EMX_DW)
            "\n -dPIDFILE_NAME   Dr.Web daemon PID-file (/var/run/drwebd.pid is default)"
            "\n -uSOCKET_FILE    Dr.Web daemon UNIX-socket file"
            "\n -cFILE_NAME      name of file for partial scanning (offset and length must be defined)"
            "\n -oOFFSET         offset in file for partial scanning"
            "\n -lLENGTH         length of file for partial scanning"
#endif /* __UNIX_DW && !EMX */
            "\n -nINETADDRESS    Dr.Web daemon Internet address (IP or domain name)"
            "\n -pPORT           Dr.Web daemon TCP port"
            "\n -tTIMEOUT        timeout (milliseconds) for communcation operations"
            "\n -h               use heuristic analyser"
            "\n -m               use spam analyser"
#if defined(__UNIX_DW)
            "\n -x               scan object as local file"
#endif /* __UNIX_DW */
            "\n -sb              show virus bases info"
            "\n -si              show daemon id string"
            "\n -sv              show daemon version"
            "\n -rv              return virus names"
            "\n -rc              return extended codes"
            "\n -rr              return report lines"
            "\n -q               quiet mode"
            "\n -w               debug mode"
            "\n"
            "\n  Return values:"
            "\n  0 - ok"
            "\n  If return code is not zero, then"
            "\n  Bit"
            "\n  0 - virus found (known, unknown, modification)"
            "\n  1 - skipped object (symlink, no regfile, skipped)"
            "\n  2 - possible mail bomb (too big compression ratio, extracted file is too big)"
            "\n  3 - daemon errors (socket & file i/o errors, timeouts, no memory)"
            "\n  4 - client errors (no memory, no daemon)"
            "\n  5 - evaluation key is used"
            "\n  6 - spam-like objects"
            "\n  7 - usage"
            "\n" );
    return DWCLIENT_USAGE;
}

