#ifndef _DW_FILTER_C
#define _DW_FILTER_C

#ifndef lint
static const char rcsid[] = "$Id: dw_filter.c,v 1.40 2001/10/15 11:04:09 cvser Exp $";
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include "dwc_lib.h"
#include "dwc_scan.h"
#include "dwc_macro.h"
#include "dw_filter.h"
#include "src/bsd_queue.h"

typedef SMFICTX         DWF_context_t;
typedef sfsistat        DWF_retcode_t;

/* static DWF_retcode_t dwf_connect ( DWF_context_t *context, char* peername, _SOCK_ADDR *peeraddr ); */
/* static DWF_retcode_t dwf_helo ( DWF_context_t *context, char* helo ); */
static DWF_retcode_t dwf_from ( DWF_context_t *context, char* from[] );
static DWF_retcode_t dwf_rcpt ( DWF_context_t *context, char* rcpt[] );
static DWF_retcode_t dwf_header ( DWF_context_t *context, char* hdr_name, char* hdr_value );
static DWF_retcode_t dwf_header_end ( DWF_context_t *context );
static DWF_retcode_t dwf_msg_chunk ( DWF_context_t *context, u_char* chunk, size_t length );
static DWF_retcode_t dwf_msg_end ( DWF_context_t *context );
static DWF_retcode_t dwf_abort ( DWF_context_t *context );
static DWF_retcode_t dwf_close ( DWF_context_t *context );

struct smfiDesc drweb_filter_desc =
{
    "DrWeb Sendmail Filter v4.26a", /* filter name */
    SMFI_VERSION,   /* version code -- do not change */
    SMFIF_ADDHDRS | SMFIF_CHGHDRS | SMFIF_ADDRCPT | SMFIF_DELRCPT | SMFIF_CHGBODY, /* flags */
    NULL,           /* connection info filter */
    NULL,           /* HELO command filter */
    dwf_from,       /* MAIL envelope sender filter */
    dwf_rcpt,       /* RCPT envelope recipient filter */
    dwf_header,     /* header filter */
    dwf_header_end, /* end of header */
    dwf_msg_chunk,  /* body block filter */
    dwf_msg_end,    /* .\r\n end of message */
    dwf_abort,      /* message or connection aborted */
    dwf_close       /* connection cleanup */
};

char  ArcDir[ PATH_MAX + 1 ] = { NULLCHAR };
char  ShareDir[ PATH_MAX + 1 ] = { NULLCHAR };
char  ArcTemplate[ PATH_MAX + 1 ] = { NULLCHAR };
char  ShareTemplate[ PATH_MAX + 1 ] = { NULLCHAR };
char  PostmasterMail[ PATH_MAX + 1 ] = { NULLCHAR };
char  NotifierMail[ PATH_MAX + 1 ] = { NULLCHAR };
int   scanOptions = DRWEBD_MAIL_MESSAGE;
int   ArcNameOffs = 0;
int   bLocal = 0;
int   bQuiet = 0;
int   bPostmasterNotify = 0;
int   bSenderNotify = 0;
int   bRcptsNotify = 0;
int   bDontSaveInfected = 0;
int   bRename = 0;

sfsistat ActionOnFatalError = SMFIS_REJECT;
sfsistat ActionOnMailbomb = SMFIS_REJECT;
sfsistat ActionOnSkipObject = SMFIS_ACCEPT;
sfsistat ActionOnError = SMFIS_REJECT;
sfsistat ActionOnSpam = SMFIS_CONTINUE;

typedef struct _DWF_rcpt_t {
    TAILQ_ENTRY( _DWF_rcpt_t )  link;
    char                       *name;
} DWF_rcpt_t;

typedef TAILQ_HEAD( _DWF_rcpt_list_t, _DWF_rcpt_t ) DWF_rcpt_list_t;

typedef struct {
    int             fd_tmp;   /* fd for temp file */
    FILE           *fs_tmp;   /* file stream for temp file */
    char           *fname_tmp;/* name of temp file */
    char           *from;     /* host that send this message */
    DWF_rcpt_list_t rcpt_list; /* recipients */
    DWF_rcpt_t     *tmp_var;  /* need for macroses */
    int             contents; /* number of headers "Content-Type" */
    int             cont_enc; /* number of headers "Content-Transfer-Encoding" */
    int             cont_dsp; /* number of headers "Content-Disposition" */
    int             mb_from_num; /* number of headers "From" */
    int             empty_from; /* flag that sets then "From" is empty */
    long            body_offset; /* offset from begin of file to data section */
} DWF_private_data_t;

/* ---------------------------------------------------------------------- */
#define ADDR( arg )  (&(arg))

/* ---------------------------------------------------------------------- */
#define get_private_data( ctx )        ((DWF_private_data_t*) smfi_getpriv((ctx)))

/* ---------------------------------------------------------------------- */
#define set_private_data( ctx, priv )  smfi_setpriv((ctx), (priv))

/* ---------------------------------------------------------------------- */
#define create_private_data( context, priv )                                \
BEGIN_MACRO                                                                 \
    (priv) = DWC_Malloc( DWF_private_data_t*, sizeof(DWF_private_data_t) ); \
    if( !(priv) )                                                           \
    {                                                                       \
        wprintLib( "can not create private data from context" );            \
                    /* it's not bug, where priv == NULL */                  \
        FAIL_EXIT( (context), (priv) );                                     \
    }                                                                       \
    TAILQ_INIT( ADDR((priv)->rcpt_list) );                                  \
    (priv)->fd_tmp = -1;                                                    \
    (priv)->fs_tmp = NULL;                                                  \
    (priv)->fname_tmp = NULL;                                               \
    (priv)->from = NULL;                                                    \
    (priv)->body_offset = 0;                                                \
    (priv)->contents = 0;                                                   \
    (priv)->cont_enc = 0;                                                   \
    (priv)->cont_dsp = 0;                                                   \
    (priv)->mb_from_num = 0;                                                \
    (priv)->empty_from = 0;                                                 \
END_MACRO

/* ---------------------------------------------------------------------- */
#define free_private_data( priv )               \
BEGIN_MACRO                                     \
    if( (priv)->fs_tmp )                        \
    {                                           \
        fflush( (priv)->fs_tmp );               \
        fclose( (priv)->fs_tmp );               \
        (priv)->fd_tmp = -1;                    \
    }                                           \
    if( (priv)->fd_tmp >= 0 )                   \
    {                                           \
        close( (priv)->fd_tmp );                \
    }                                           \
    if( (priv)->fname_tmp )                     \
    {                                           \
        unlink( (priv)->fname_tmp );            \
        /* asv: allocated by strdup */          \
        free( (priv)->fname_tmp );              \
    }                                           \
    if( (priv)->from )                          \
    {   /* asv: allocated by strdup */          \
        free( (priv)->from );                   \
    }                                           \
    while( ((priv)->tmp_var = TAILQ_FIRST( ADDR((priv)->rcpt_list) )) != NULL )\
    {   /* asv: unlink & free recipients */     \
        TAILQ_REMOVE( ADDR((priv)->rcpt_list), (priv)->tmp_var, link );\
        /* asv: allocated by strdup */          \
        free( (priv)->tmp_var->name );          \
        DWC_Free( (priv)->tmp_var );            \
    }                                           \
    DWC_Free( (priv) );                         \
END_MACRO



/* ---------------------------------------------------------------------- */
#define FAIL_EXIT( context, priv )              \
BEGIN_MACRO                                     \
    if( (priv) )                                \
    {                                           \
        free_private_data( (priv) );            \
    }                                           \
    if( set_private_data( (context), NULL ) != MI_SUCCESS )    \
    {                                                          \
        wprintLib( "%s(%d) - FATAL ERROR: can not null private data", __FILE__, __LINE__ );\
    }                                           \
    return ActionOnFatalError;                  \
END_MACRO

/* ---------------------------------------------------------------------- */
#define PROLOGUE( context, priv )                                   \
BEGIN_MACRO                                                         \
    (priv) = get_private_data( (context) );                         \
    if( !(priv) )                                                   \
    {                                                               \
        wprintLib( "%s(%d) - FATAL ERROR: can not extract private data from context", __FILE__, __LINE__ );   \
                      /* it's not bug, where priv == NULL */        \
        FAIL_EXIT( (context), (priv) );                             \
    }                                                               \
END_MACRO

/* ----------------------------------------------------------------------- */
int set_action( char opt, char* all_opt, sfsistat* action )
{
    if( !all_opt || !action )
    {
        return 0;
    }
    switch( opt )
    {
        case 'a':
            if( strchr( all_opt, 'a' ) == NULL )
                return 0;
            *action = SMFIS_ACCEPT;
            break;
        case 'c':
            if( strchr( all_opt, 'c' ) == NULL )
                return 0;
            *action = SMFIS_CONTINUE;
            break;
        case 'd':
            if( strchr( all_opt, 'd' ) == NULL )
                return 0;
            *action = SMFIS_DISCARD;
            break;
        case 'r':
            if( strchr( all_opt, 'r' ) == NULL )
                return 0;
            *action = SMFIS_REJECT;
            break;
        case 't':
            if( strchr( all_opt, 't' ) == NULL )
                return 0;
            *action = SMFIS_TEMPFAIL;
            break;
        default:
            return 0;
    }
    return 1;
}

/* ----------------------------------------------------------------------- */
char get_action( sfsistat action )
{
    char  ret = 'U';

    switch( action )
    {
     case SMFIS_ACCEPT:
        ret = 'A';
        break;
     case SMFIS_DISCARD:
        ret = 'D';
        break;
     case SMFIS_TEMPFAIL:
        ret = 'T';
        break;
     case SMFIS_REJECT:
        ret = 'R';
        break;
     case SMFIS_CONTINUE:
        ret = 'C';
        break;
     default:
        break;
    }
    return ret;
}

/* ---------------------------------------------------------------------- */
int preinit_filter( void )
{
    char  *tmp_name = NULL;
    char  *arc_name = NULL;
    int    error, fd = -1;
    int    f_tmp = 0;
    int    f_arc = 0;
    char   test_buf[] = { 'T', 'E', 'S', 'T', '\000' };

    if( ArcDir[0] == NULLCHAR )
    {
        wprintLib( "Directory for infected files not specified, such files would be discarded" );
        bDontSaveInfected = 1;
    }
    if( access( ShareDir, W_OK | X_OK ) < 0 )
    {
        wprintLib( "Directory %s for temporary files is unaccessible - %m.\n", ShareDir );
        return -1;
    }
    if( snprintf( ShareTemplate, sizeof(ShareTemplate)-1,
                  "%s/drweb.tmp.XXXXXX", ShareDir ) < 0 )
    {
        wprintLib( "Path is too long for directory proposed for temporary files.\n" );
        return -1;
    }
    if( !bDontSaveInfected && (access( ArcDir, W_OK | X_OK ) < 0) )
    {
        wprintLib( "Directory %s for infected files is unaccessible - %m.\n", ArcDir );
        return -1;
    }
    ArcNameOffs = strlen(ArcDir) + sizeof(char);
    if( !bDontSaveInfected 
        && 
        (snprintf( ArcTemplate, sizeof(ArcTemplate)-1,
                  "%s/archive.msg.XXXXXX", ArcDir ) < 0) )
    {
        wprintLib( "Path is too long for directory for infected files.\n" );
        return -1;
    }
    tmp_name = strdup( ShareTemplate );
    if( !tmp_name )
    {
        wprintLib( "No memory for test moving function.\n" );
        error = -1;
        goto error_exit;
    }
    fd = mkstemp( tmp_name );
    if( fd < 0 )
    {
        wprintLib( "Can not create temporary file %s - (%u) %s.\n", tmp_name, errno, strerror(errno) );
        error = -1;
        goto error_exit;
    }
    f_tmp = 1;
    if( write( fd, test_buf, sizeof( test_buf ) ) < 0 )
    {
        wprintLib( "Can not write to temporary file - (%u) %s.\n", errno, strerror(errno) );
        error = -1;
        goto error_exit;
    }
    if( close( fd ) < 0 )
    {
        wprintLib( "Can not close temporary file - (%u) %s.\n", errno, strerror(errno) );
        error = -1;
        goto error_exit;
    }
    if( bDontSaveInfected )
    {
        error = 0;
        goto error_exit;
    }
    arc_name = strdup( ArcTemplate );
    if( !arc_name )
    {
        wprintLib( "No memory for test moving function.\n" );
        error = -1;
        goto error_exit;
    }
    fd = mkstemp( arc_name );
    if( fd < 0 )
    {
        wprintLib( "Can not create archive file %s - (%u) %s.\n", arc_name, errno, strerror(errno) );
        error = -1;
        goto error_exit;
    }
    f_arc = 1;
    if( write( fd, test_buf, sizeof( test_buf ) ) < 0 )
    {
        wprintLib( "Can not write to archive file - (%u) %s.\n", errno, strerror(errno) );
        error = -1;
        goto error_exit;
    }
    if( close( fd ) < 0 )
    {
        wprintLib( "Can not close archive file - (%u) %s.\n", errno, strerror(errno) );
        error = -1;
        goto error_exit;
    }
    if( rename( tmp_name, arc_name ) < 0 )
    {
        if( errno == EXDEV )
        {
            wprintLib( "TMP_DIR and INFECTED_DIR are located on different partitions, so we can use only slow moving routine." );
            bRename = 0;
        }
        else
        {
            wprintLib( "Can not rename temporary file to archive file - (%u) %s.\n", errno, strerror(errno) );
            error = -1;
            goto error_exit;
        }
    }
    else
    {
        bRename = 1;
    }
    error = 0;
  error_exit:
    if( tmp_name )
    {
        if( f_tmp )
            unlink( tmp_name );
        free( tmp_name );
    }
    if( arc_name )
    {
        if( f_arc )
            unlink( arc_name );
        free( arc_name );
    }
    return error;
}

/* ---------------------------------------------------------------------- */
#ifndef SSIZE_MAX
#define SSIZE_MAX   4096
#endif /* SSIZE_MAX */
#define BUF_LEN  ( SSIZE_MAX > 0x1000 ? 0x1000 : SSIZE_MAX )
static int move_file( const char* oldpath, const char* newpath )
{
    char*       buf;
    int         len, towrite, written, error, save_errno, src_fd, dst_fd, wrb;
    struct stat st;

    save_errno = 0;
    dst_fd = src_fd = -1;
    buf = DWC_Malloc( char*, BUF_LEN );
    if( !buf )
    {
        save_errno = ENOMEM;
        wprintLib( "Moving file: can not allocate buffer for moving.\n" );
        error = -1;
        goto error_exit;
    }
    if( stat( oldpath, &st ) < 0 )
    {
        save_errno = errno;
        wprintLib( "Moving file: temporary file stat() error.\n" );
        error = -1;
        goto error_exit;
    }
    len = st.st_size;
    dst_fd = open( newpath, O_CREAT | O_WRONLY, st.st_mode );
    if( dst_fd < 0 )
    {
        save_errno = errno;
        wprintLib( "Moving file: can not open archive file.\n" );
        error = -1;
        goto error_exit;
    }
    src_fd = open( oldpath, O_RDONLY );
    if( src_fd < 0 )
    {
        save_errno = errno;
        wprintLib( "Moving file: can not open temporary file.\n" );
        error = -1;
        goto error_exit;
    }
    if( lseek( src_fd, 0, SEEK_SET ) < 0 )
    {
        save_errno = errno;
        wprintLib( "Moving file: can not seek to begin of temporary file.\n" );
        error = -1;
        goto error_exit;
    }
    if( lseek( dst_fd, 0, SEEK_SET ) < 0 )
    {
        save_errno = errno;
        wprintLib( "Moving file: can not seek to begin of archive file.\n" );
        error = -1;
        goto error_exit;
    }
    while( len )
    {
        towrite = read( src_fd, buf, BUF_LEN );
        if( towrite < 0 )
        {
            save_errno = errno;
            wprintLib( "Moving file: can not read temporary file.\n" );
            error = -1;
            goto error_exit;
        }
        if( towrite == 0 )
        {
            /* end-of-file */
            break;
        }
        written = 0;
        do
        {
            wrb = write( dst_fd, buf + written, towrite - written );
            if( wrb < 0 )
            {
                save_errno = errno;
                wprintLib( "Moving file: archive file writing error (%d:%d) - %s.\n", towrite, written, strerror(errno) );
                error = -1;
                goto error_exit;
            }
            written += wrb;
        } while( written < towrite );
        len -= towrite;
    }

    error = 0;
  error_exit:
    if( src_fd >= 0 )
    {
        if( close( src_fd ) < 0 )
        {
            save_errno = errno;
            wprintLib( "Moving file: cannot close src-file.\n" );
            error = -1;
        }
    }
    if( dst_fd >= 0 )
    {
        if( close( dst_fd ) < 0 )
        {
            save_errno = errno;
            wprintLib( "Moving file: cannot close dst-file.\n" );
            error = -1;
        }
    }
    DWC_Free( buf );
    errno = save_errno;
    return error;
}

/* ---------------------------------------------------------------------- */
static int reply_file( DWF_context_t *context, const char* path, long offset )
{
    char*       buf;
    int         len, max_len, readed, written, error, save_errno, src_fd;
    struct stat st;

    buf = NULL;
    src_fd = -1;
    save_errno = 0;
    if( stat( path, &st ) < 0 )
    {
        save_errno = errno;
        dprintLib( "can not stat file for trial reply.\n" );
        error = -1;
        goto error_exit;
    }
    len = st.st_size - offset;
    buf = DWC_Malloc( char*, len );
    if( buf == NULL )
    {
        save_errno = errno;
        dprintLib( "can not allocate memory for trial reply.\n" );
        error = -1;
        goto error_exit;
    }
    src_fd = open( path, O_RDONLY );
    if( src_fd < 0 )
    {
        save_errno = errno;
        dprintLib( "can not open file for trial reply.\n" );
        error = -1;
        goto error_exit;
    }
    if( lseek( src_fd, offset, SEEK_SET ) < 0 )
    {
        save_errno = errno;
        dprintLib( "can not seek to offset=%d file for trial reply.\n", offset );
        error = -1;
        goto error_exit;
    }
    written = 0;
    while( written < len )
    {
        if( (len - written) > BUF_LEN )
            max_len = BUF_LEN;
        else
            max_len = len - written;
        readed = read( src_fd, buf + written, max_len );
        if( readed < 0 )
        {
            save_errno = errno;
            dprintLib( "can not read file on trial reply.\n" );
            error = -1;
            goto error_exit;
        }
        if( readed == 0 )
        {
            /* end-of-file */
            break;
        }
        written += readed;
    }

    if( smfi_replacebody( context, buf, len ) != MI_SUCCESS )
    {
        dprintLib( "can not replace body on trial reply.\n" );
    }
    error = 0;
  error_exit:
    DWC_Free( buf ); /* buf = NULL after that */
    if( src_fd >= 0 )
    {
        if( close( src_fd ) < 0 )
        {
            save_errno = errno;
            dprintLib( "can not close message file after trial-msg.\n" );
            error = -1;
        }
    }
    errno = save_errno;
    return error;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  IN: char *peername;
 *     Host domain name, as determined by a reverse lookup on the host
 *     address.
 *  IN: _SOCK_ADDR *peeraddr;
 *     Host address, as determined by a getpeername call on the SMTP
 *     socket.
 *  NOTE: Invoked on each connection
 * ----------------------------------------------------------------------- */
/*
DWF_retcode_t dwf_connect ( DWF_context_t *context, char* peername, _SOCK_ADDR *peeraddr )
{
    if( peername )
        dprintLib( "connect with %s", peername );
    else
        dprintLib( "connect with unknown" );
    return SMFIS_CONTINUE;
}
*/
/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  IN: char *helo;
 *     Value passed to HELO/EHLO command, which should be the domain
 *     name of the sending host (but is, in practice, anything the
 *     sending host wants to send).
 *  NOTE: Invoked on SMTP HELO/EHLO command
 * ----------------------------------------------------------------------- */
/*
DWF_retcode_t dwf_helo ( DWF_context_t *context, char* helo )
{
    if( helo && helo[0] )
        dprintLib( "HELO %s" helo );
    else
        dprintLib( "HELO <UNKNOWN>" );
    return SMFIS_CONTINUE;
}
*/
/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  IN: char **from;
 *     Null-terminated SMTP command arguments. from[0] is guaranteed
 *     to be the sender address. Later arguments are the ESMTP
 *     arguments.
 *  NOTE: Invoked on envelope from
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_from ( DWF_context_t *context, char* from[] )
{
    DWF_private_data_t  *priv;

    if( from && from[0] )
        dprintLib( "envelope from %s", from[0] );
    else
        dprintLib( "envelope from <NULL>" );

    create_private_data( context, priv ); /* macro - return if error */
    /* save originator of message */
    if( from && from[0] && (strlen(from[0]) > 0) && strcmp(from[0], "<>") )
    {
        priv->empty_from = 0;
        priv->from = strdup( from[0] );
    }
    else
    {
        /* empty or <> From: - looks like a spam */
        priv->empty_from = 1;
        if( ActionOnSpam == SMFIS_CONTINUE )
        {
            priv->from = strdup( "<>" );
        }
        else
        {
            wprintLib( "block processing from <>" );
            free_private_data( priv );
            if( set_private_data( context, NULL ) != MI_SUCCESS )
            {
                wprintLib( "%s(%d) - FATAL ERROR: can't store private data", __FILE__, __LINE__ );
                return ActionOnFatalError;
            }
            return ActionOnSpam;
        }
    }
    if( !priv->from )
    {
        wprintLib( "%s(%d) - no memory for private data", __FILE__, __LINE__ );
        FAIL_EXIT( context, priv );
    }
    wprintLib( "start processing from %s", priv->from );
    priv->fname_tmp = strdup( ShareTemplate );
    if( !priv->fname_tmp )
    {
        wprintLib( "%s(%d) - no memory for private data", __FILE__, __LINE__ );
        FAIL_EXIT( context, priv );
    }
    priv->fd_tmp = mkstemp( priv->fname_tmp );
    if( priv->fd_tmp < 0 )
    {
        wprintLib( "can not create tempoary file - %m" );
        FAIL_EXIT( context, priv );
    }
    priv->fs_tmp = fdopen( priv->fd_tmp, "w+" );
    if( !priv->fs_tmp )
    {
        wprintLib( "can not open stream for temporary file - %m" );
        FAIL_EXIT( context, priv );
    }
    if( !priv->from )
    {
        wprintLib( "no memory for private data" );
        FAIL_EXIT( context, priv );
    }
    /* fprintf( priv->fs_tmp, "Return-Path: %s\r\n", priv->from ); */
    if( set_private_data( context, priv ) != MI_SUCCESS )
    {
        wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
        FAIL_EXIT( context, priv );
    }
    return SMFIS_CONTINUE;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  IN: char **rcpt;
 *     Null-terminated SMTP command arguments. rcpt[0] is guaranteed
 *     to be the recipient address. Later arguments are the ESMTP
 *     arguments.
 *  NOTE: Invoked on each envelope recipient
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_rcpt ( DWF_context_t *context, char* rcpt[] )
{
    /* int                 i; */
    DWF_private_data_t  *priv;

    dprintLib( "envelope to %s", rcpt[0] );

    if( !bRcptsNotify && (strlen(rcpt[0]) > 0) )
    {
        PROLOGUE( context, priv );
        /*
        fprintf( priv->fs_tmp, "Envelope-To:" );
        for( i = 0; rcpt[i] != NULL; ++i )
        {
            fprintf( priv->fs_tmp, " %s,", rcpt[i] );
        }
        fprintf( priv->fs_tmp, "\r\n" );
        */
        priv->tmp_var = DWC_Malloc( DWF_rcpt_t*, sizeof(DWF_rcpt_t) );
        if( !priv->tmp_var )
        {
            wprintLib( "%s(%d) - can not allocate memory", __FILE__, __LINE__ );
            FAIL_EXIT( context, priv );
        }
        priv->tmp_var->name = strdup( rcpt[0] );
        if( !priv->tmp_var->name )
        {
            wprintLib( "%s(%d) - can not allocate memory", __FILE__, __LINE__ );
            DWC_Free( priv->tmp_var );
            FAIL_EXIT( context, priv );
        }
        TAILQ_INSERT_HEAD( ADDR(priv->rcpt_list), priv->tmp_var, link );
        if( set_private_data( context, priv ) != MI_SUCCESS )
        {
            wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
            FAIL_EXIT( context, priv );
        }
    }
    return SMFIS_CONTINUE;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  IN: char *hdr_name;
 *     Header field name
 *  IN: char *hdr_value;
 *     Header field value
 * NOTE: Invoked on each message header. The content of the header may
 *       have folded white space (that is, multiple lines with
 *       following white space) included.
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_header ( DWF_context_t *context, char* hdr_name, char* hdr_value )
{
    DWF_private_data_t  *priv;

    dprintLib( "header - %s:%s", hdr_name, hdr_value );

    if( !hdr_name )
    {
        return SMFIS_CONTINUE;
    }
    PROLOGUE( context, priv );
    if( strncmp( hdr_name, "Content-Type", strlen("Content-Type") ) == 0 )
    {
        ++priv->contents;
    }
    if( strncmp( hdr_name, "Content-Transfer-Encoding", strlen("Content-Transfer-Encoding") ) == 0 )
    {
        ++priv->cont_enc;
    }
    if( strncmp( hdr_name, "Content-Disposition", strlen("Content-Disposition") ) == 0 )
    {
        ++priv->cont_dsp;
    }
    if( strncmp( hdr_name, "From", strlen("From") ) == 0 )
    {
        ++priv->mb_from_num;
    }
    if( hdr_value )
    {
        fprintf( priv->fs_tmp, "%s: %s\r\n", hdr_name, hdr_value );
    }
    else
    {
        fprintf( priv->fs_tmp, "%s: <NULL>\r\n", hdr_name );
    }
    if( set_private_data( context, priv ) != MI_SUCCESS )
    {
        wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
        FAIL_EXIT( context, priv );
    }
    return SMFIS_CONTINUE;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  NOTE: Invoked at end of header
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_header_end ( DWF_context_t *context )
{
    DWF_private_data_t  *priv;

    dprintLib( "end of headers section" );

    PROLOGUE( context, priv );
    fprintf( priv->fs_tmp, "\r\n" );
    if( bLocal )
    {
        priv->body_offset = ftell( priv->fs_tmp );
    }
    if( set_private_data( context, priv ) != MI_SUCCESS )
    {
        wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
        FAIL_EXIT( context, priv );
    }
    return SMFIS_CONTINUE;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  IN: u_char *bodyp;
 *     Pointer to chuck of message body
 *  IN: size_t bodylen;
 *     Length of body data
 *  NOTE: Invoked for each body chunk. There may be multiple body
 *        chunks passed to the filter. End-of-line passed as received
 *        from SMTP
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_msg_chunk ( DWF_context_t *context, u_char* chunk, size_t length )
{
    DWF_private_data_t  *priv;

    dprintLib( "message chunk - %d bytes", length );

    PROLOGUE( context, priv );
    if( fwrite( chunk, length, 1, priv->fs_tmp ) != 1 )
    {
        wprintLib( "can not save message body to temporary file - %m" );
        FAIL_EXIT( context, priv );
    }
    if( set_private_data( context, priv ) != MI_SUCCESS )
    {
        wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
        FAIL_EXIT( context, priv );
    }
    return SMFIS_CONTINUE;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  NOTE: Invoked at end of message. This routine can perform special
 *        operations such as modifying the message header, body, or envelope.
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_msg_end ( DWF_context_t *context )
{
    int                 fd, ret, len, scan_type, i, status, loglen;
    DWF_private_data_t  *priv;
    char                *name = NULL;
    char                NotifyMsg[4096];
    char                ReportMsg[1024];
    char                ArchiveMsg[1024];
    char                *plog;
    int                 *plen;
    sfsistat            exit_code;
    int                 virus = 0;

    dprintLib( "end of message chunks" );

    PROLOGUE( context, priv );
    if( priv->fs_tmp )
    {
        fflush( priv->fs_tmp );
        fclose( priv->fs_tmp );
        priv->fd_tmp = -1;
        priv->fs_tmp = NULL;
    }
    if( bLocal )
    {
        scan_type = DWL_FT_DISKFILE;
    }
    else
    {
        scan_type = DWL_FT_NORMALFILE;
    }
    if( scanOptions & DRWEBD_RETURN_VIRUSES )
    {
        loglen = sizeof(ReportMsg);
        plog = ReportMsg;
        plen = &loglen;
    }
    else
    {
        plog = NULL;
        plen = NULL;
    }
    dprintLib( "start scanning message from %s ...", priv->from );
    ret = scanFile( priv->fname_tmp, strlen(priv->fname_tmp),
                    scan_type, scanOptions, &status, plog, plen );
        /* asv - after scan we (currently) dont make tempfail (todo) */
    if( ret != DWL_NOERROR )
    {
        wprintLib( "scan file failed (%s) - possible problem with client or daemon", getLibError( ret ) );
        exit_code = ActionOnError;
    }
    else if( status & DERR_VIRUS_MASK )
    {   /* viruses found */
        virus = 1;
        wprintLib( "message sent by %s is infected.\n", priv->from );
        if( !bDontSaveInfected )
        {
            name = strdup( ArcTemplate );
            if( !name )
            {
                wprintLib( "%s(%d) - no memory for temporary data", __FILE__, __LINE__ );
                FAIL_EXIT( context, priv );
            }
            fd = mkstemp( name );
            if( fd < 0 )
            {
                wprintLib( "can not create file for store infected mail - %m" );
                free( name );
                FAIL_EXIT( context, priv );
            }
            if( close( fd ) < 0 )
            {
                wprintLib( "can not close file for store infected mail - %m" );
                free( name );
                FAIL_EXIT( context, priv );
            }
            if( bRename )
            {
                ret = rename( priv->fname_tmp, name );
            }
            else
            {
                ret = move_file( priv->fname_tmp, name );
            }
            if( ret < 0 )
            {
                wprintLib( "can not move file to archive - %m" );
                free( name );
                FAIL_EXIT( context, priv );
            }

            wprintLib( "infected message sent by %s has been stored in archive %s", priv->from, name );
            
            if( snprintf( ArchiveMsg, sizeof(ArchiveMsg),
                            "Original message was stored in the archive.\r\n"
                            "To receive original message please contact postmaster:\r\n%s\r\n\r\n"
                            "Archive record: %s\r\n", NotifierMail, name + ArcNameOffs ) < 0 )
            {
                wprintLib( "can not create archive message, skipped" );
                ArchiveMsg[0] = NULLCHAR;
            }
        }            
        else
        {
            ArchiveMsg[0] = NULLCHAR;
        }

        if( priv->contents > 0 )
        {
            if( smfi_chgheader( context, "Content-Type", 1, "text/plain; charset=us-ascii" ) != MI_SUCCESS )
            {
                wprintLib( "can not change content type of mail" );
                free( name );
                FAIL_EXIT( context, priv );
            }
            --priv->contents;
            for( i = 0; i < priv->contents; ++i )
            {
                if( smfi_chgheader( context, "Content-Type", i+2, "" ) != MI_SUCCESS )
                {
                    wprintLib( "can not change content type of mail" );
                    free( name );
                    FAIL_EXIT( context, priv );
                }
            }
        }
        else
        {
            if( smfi_addheader( context, "Content-Type", "text/plain; charset=us-ascii" ) != MI_SUCCESS )
            {
                wprintLib( "can not set content type of mail" );
                free( name );
                FAIL_EXIT( context, priv );
            }
        }
        for( i = 0; i < priv->cont_enc; ++i )
        {
            if( smfi_chgheader( context, "Content-Transfer-Encoding", i+1, "" ) != MI_SUCCESS )
            {
                wprintLib( "can not delete content encoding in mail" );
                free( name );
                FAIL_EXIT( context, priv );
            }
        }
        for( i = 0; i < priv->cont_dsp; ++i )
        {
            if( smfi_chgheader( context, "Content-Disposition", i+1, "" ) != MI_SUCCESS )
            {
                wprintLib( "can not delete content disposition in mail" );
                free( name );
                FAIL_EXIT( context, priv );
            }
        }        
        if( plog )
        {
            len = snprintf( NotifyMsg, sizeof(NotifyMsg),
                            "Message sent from %s\r\n"
                            "infected by virus and has not been delivered.\r\n"
                            "Viruses:\r\n%s\r\n"
                            "%s"
                            "\tAntivirus service provided by Dr.Web Daemon (www.sald.com)\r\n",
                            priv->from, plog, ArchiveMsg );
        }
        else
        {
            len = snprintf( NotifyMsg, sizeof(NotifyMsg),
                            "Message sent from %s\r\n"
                            "infected by virus and has not been delivered.\r\n"
                            "%s"
                            "\tAntivirus service provided by Dr.Web Daemon (www.sald.com)\r\n",
                            priv->from, ArchiveMsg );
        }
        free( name );
        if( len < 0 )
        {
            wprintLib( "can't create notify message" );
            FAIL_EXIT( context, priv );
        }
        if( smfi_chgheader( context, "From", 1, NotifierMail ) != MI_SUCCESS )
        {
            wprintLib( "can not add sender" );
            FAIL_EXIT( context, priv );
        }
        for( i = 0; i < priv->mb_from_num; ++i )
        {
            if( smfi_chgheader( context, "From", i+2, "" ) != MI_SUCCESS )
            {
                wprintLib( "can not change aux <from>" );
                FAIL_EXIT( context, priv );
            }
        }
        if( !bRcptsNotify )
        {
            TAILQ_FOREACH( priv->tmp_var, ADDR(priv->rcpt_list), link )
            {
                if( smfi_delrcpt( context, priv->tmp_var->name ) != MI_SUCCESS )
                {
                    wprintLib( "can not delete recipients" );
                    FAIL_EXIT( context, priv );
                }                
            }
        }
        if( bSenderNotify && !priv->empty_from )
        {
            if( smfi_addrcpt( context, priv->from ) != MI_SUCCESS )
            {
                wprintLib( "can not add sender to recipients" );
                FAIL_EXIT( context, priv );
            }
        }
        if( bPostmasterNotify )
        {
            if( smfi_addrcpt( context, PostmasterMail ) != MI_SUCCESS )
            {
                wprintLib( "can not add admin to recipients" );
                FAIL_EXIT( context, priv );
            }
        }
        if( smfi_replacebody( context, NotifyMsg, len ) != MI_SUCCESS )
        {
            wprintLib( "can not replace message" );
            FAIL_EXIT( context, priv );
        }
        exit_code = SMFIS_ACCEPT;
    }
    else if( status & DERR_SPAM )
    {
        wprintLib( "message sent by %s looks like a spam.\n", priv->from );
        if( !bDontSaveInfected )
        {
            name = strdup( ArcTemplate );
            if( !name )
            {
                wprintLib( "%s(%d) - no memory for temporary data", __FILE__, __LINE__ );
                FAIL_EXIT( context, priv );
            }
            fd = mkstemp( name );
            if( fd < 0 )
            {
                wprintLib( "can not create file for store spam mail - %m" );
                free( name );
                FAIL_EXIT( context, priv );
            }
            if( close( fd ) < 0 )
            {
                wprintLib( "can not close file for store spam mail - %m" );
                free( name );
                FAIL_EXIT( context, priv );
            }
            if( bRename )
            {
                ret = rename( priv->fname_tmp, name );
            }
            else
            {
                ret = move_file( priv->fname_tmp, name );
            }
            free( name );
            if( ret < 0 )
            {
                wprintLib( "can not move file to archive - %m" );
                FAIL_EXIT( context, priv );
            }
            wprintLib( "message sent by %s has been stored in archive %s", priv->from, name );        
        }
        exit_code = ActionOnSpam;
        if( exit_code == SMFIS_CONTINUE )
            exit_code = SMFIS_ACCEPT;
    }
    else if( status & DERR_MAIL_BOMB )
    {
        wprintLib( "message sent by %s possible contains a mailbomb.\n", priv->from );
        exit_code = ActionOnMailbomb;
    }
    else if( status & DERR_SKIP_OBJECT )
    {
        wprintLib( "message sent by %s contains an unchecked objects.\n", priv->from );
        exit_code = ActionOnSkipObject;
    }
    else if( (status != DERR_NOERROR) && (status != DERR_EVAL_KEY) )
    {
        wprintLib( "daemon return error (%s) - possible problem with daemon or file", getDaemonCode( status ) );
        exit_code = ActionOnError;
    }
    else
    {
        wprintLib( "message sent by %s is okay.\n", priv->from );
        exit_code = SMFIS_ACCEPT;
    }
    if( bLocal && (virus == 0) && (exit_code == SMFIS_ACCEPT) && (status & DERR_EVAL_KEY) && (priv->body_offset > 0) )
    {
        reply_file( context, priv->fname_tmp, priv->body_offset );
    }
    dprintLib( "processing message from %s completed", priv->from );
    free_private_data( priv );
    if( set_private_data( context, NULL ) != MI_SUCCESS )
    {
        wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
        return ActionOnFatalError;
    }
    return exit_code;
}

/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  NOTE: Invoked at end of the connection. This is called on close
 *        even if the previous mail transaction was aborted.
 *
 ----------------------------------------------------------------------- */
DWF_retcode_t dwf_close ( DWF_context_t *context )
{
    DWF_private_data_t *priv;

    dprintLib( "close SMTP session" );

    priv = get_private_data( context );
    if( priv )
    {
        free_private_data( priv );
        if( set_private_data( context, NULL ) != MI_SUCCESS )
        {
            wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
            return ActionOnFatalError;
        }
        return SMFIS_ACCEPT;
    }
    else
    {
        return SMFIS_REJECT;
    }
}


/* -----------------------------------------------------------------------
 *  IN: SMFICTX *context;
 *     Context of current filtering session
 *  NOTE: Invoked if message is aborted outside of the control of the
 *        filter, for example, if the SMTP sender issues an RSET
 *        command. If dwf_abort is called, msg_end will not be called
 *        and vice versa.
 * ----------------------------------------------------------------------- */
DWF_retcode_t dwf_abort ( DWF_context_t *context )
{
    DWF_private_data_t *priv;

    dprintLib( "SMTP session abort or message processing" );

    priv = get_private_data( context );
    if( priv )
    {
        wprintLib( "message from %s is aborted", priv->from );
        free_private_data( priv );
        if( set_private_data( context, NULL ) != MI_SUCCESS )
        {
            wprintLib( "%s(%d) - FATAL ERROR: can not store private data", __FILE__, __LINE__ );
            return ActionOnFatalError;
        }
    }
    return SMFIS_REJECT;
}

#endif /* _DW_FILTER_C */
