/*--------------------------------------------------------------------------
 Dr.Web daemon client for Sendmail Mail Filter API
                                       ID Antivirus Lab http://www.drweb.ru
                                        http://www.drweb.ru/linux-bsd/forum
                                              http://www.drweb.ru/linux-bsd
----------------------------------------------------------------------------
 Acknowledgements:
 - Andrey Chernov <ache@nagual.pp.ru>
 - Anatoly A. Orehovsky <tolik@asplinux.ru>
--------------------------------------------------------------------------*/

#ifndef lint
static const char rcsid[] = "$Id: dwf_main.c,v 1.28 2001/10/15 11:04:09 cvser Exp $";
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <pwd.h>

#include "dw_filter.h"
#include "dwc_lib.h"
#include "dwc_macro.h"

#define DWCLIENT_NOERROR        (0)     /* 0x00000000 */
#define DWCLIENT_EXPAND         (1<<31) /* 0x80000000 */
#define DWCLIENT_DEMO_USAGE     (1<<30) /* 0x40000000 */
#define DWCLIENT_NO_MEMORY      (1<<29) /* 0x20000000 */
#define DWCLIENT_LIB_ERR        (1<<28) /* 0x10000000 */
#define DWCLIENT_SYS_ERR        (1<<27) /* 0x08000000 */
#define DWCLIENT_PARAMETERS     (1<<26) /* 0x04000000 */

static char UserName[ 255 ] = { NULLCHAR };
static char PidFile[ PATH_MAX + 1 ] = { NULLCHAR };
static char DaemonUnix[ PATH_MAX + 1 ] = { NULLCHAR };
static char DaemonInet[ PATH_MAX + 1 ] = { NULLCHAR };
static int  DaemonPort = 0;
static char SendmailUnix[ PATH_MAX + 1 ] = { NULLCHAR };
static char SendmailInet[ PATH_MAX + 1 ] = { NULLCHAR };
static int  SendmailPort = 0;
static int  bDUnix = 0;
static int  bDInet = 0;
static int  bSMUnix = 0;
static int  bSMInet = 0;
static int  bDebug = 0;
static int  bCache = 0;
static int  Timeout = 0;

static int   parseCmdLine( int argc, char ** argv );
static pid_t daemonize( void );
static int   setuser( void );
static int   usage( void );

/* ----------------------------------------------------------------------- */
int main( int argc, char *argv[] )
{
    int     retcode = DWCLIENT_NOERROR;
    int     libcode = DWL_NOERROR;
    pid_t   pid;
    char    Buffer[ PATH_MAX + 1 ] = { NULLCHAR };

    if( !parseCmdLine( argc, argv ) )
    {
        return usage();
    }
    outputLib( bQuiet, bDebug, 1 );
    retcode = setuser();
    if( retcode != DWCLIENT_NOERROR )
    {
        wprintLib("Can not set new user.\n");
        return DWCLIENT_LIB_ERR;
    }
    pid = daemonize();
    if( pid > 0 )
    {
        return DWCLIENT_NOERROR;
    }
    else if( pid < 0 )
    {
        wprintLib("Can not daemonize.\n");
        return DWCLIENT_SYS_ERR;
    }
    if( bDInet )
    {
        libcode = startupLib( DWL_TYPE_TCP_SOCKET, Timeout, DaemonInet, DaemonPort, bCache );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can not initialize library with TCP/IP transport.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    else if( bDUnix )
    {
        libcode = startupLib( DWL_TYPE_UNIX_SOCKET, Timeout, DaemonUnix );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can not initialize library with UNIX sockets transport.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    else if( PidFile[0] != NULLCHAR )
    {
        libcode = startupLib( DWL_TYPE_READPIDFILE, Timeout, PidFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can not initialize library (pidfile reading error).\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    else
    {
        dprintLib("Daemon address is not defined, defaults will be used.\n");
        libcode = startupLib( DWL_TYPE_DEFAULT, Timeout );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Library initialization error.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    if( bSMUnix )
    {
        if( snprintf( Buffer, sizeof(Buffer)-1, "local:%s", SendmailUnix ) < 0 )
        {
            wprintLib("Can not format address for initialize sendmail filter.\n");
            retcode = DWCLIENT_SYS_ERR;
            goto error_exit;
        }
        Buffer[sizeof(Buffer)-1] = NULLCHAR;
    }
    else /* bSMInet - other alternative absent - it's guaranted by ParseCmdLine */
    {
        if( snprintf( Buffer, sizeof(Buffer)-1, "inet:%d@%s", SendmailPort, SendmailInet ) < 0 )
        {
            wprintLib("Can not format address for initialize sendmail filter.\n");
            retcode = DWCLIENT_SYS_ERR;
            goto error_exit;
        }
        Buffer[sizeof(Buffer)-1] = NULLCHAR;
    }
    /* next check access, so we should drop privilegies before execute it */
    if( preinit_filter() < 0 )
    {
        wprintLib( "Any parameters contain invalid values.\n" );
        goto error_exit;
    }
    if( smfi_setconn( Buffer ) != MI_SUCCESS )
    {
        wprintLib( "Can not set address for sendmail filter.\n" );
        retcode = DWCLIENT_SYS_ERR;
        goto error_exit;
    }
    dprintLib( "connected to %s", Buffer );
    if( smfi_register( drweb_filter_desc ) != MI_SUCCESS )
    {
        wprintLib( "Can not register filter.\n" );
        retcode = DWCLIENT_SYS_ERR;
        goto error_exit;
    }
    wprintLib( "Dr.Web sendmail filter started ..." );
    dprintLib( "Actions: spam=%c, skip=%c, bomb=%c, error=%c, fatal=%c.\n", get_action(ActionOnSpam), get_action(ActionOnSkipObject), get_action(ActionOnMailbomb), get_action(ActionOnError), get_action(ActionOnFatalError) );
    if( smfi_main() != MI_SUCCESS )
    {
        wprintLib( "Dr.Web sendmail filter returned abnormally ...\n" );
        retcode = DWCLIENT_SYS_ERR;
    }
    wprintLib( "Dr.Web sendmail filter shutdown in progress..." );
  error_exit:
    cleanupLib();
    if( (retcode & ~DERR_MASK ) != 0 ) /* have local errors */
    {
        if( retcode & DWCLIENT_LIB_ERR )
        {
            dprintLib("Results: dwclient library error %d (%s)\n", libcode, getLibError( libcode ) );
        }
    }
    else
    {
        if( retcode != DERR_NOERROR ) /* scan not ok - errors or viruses */
        {
            dprintLib("Results: daemon return code %#x (%s)\n", retcode, getDaemonCode( retcode ) );
        }
    }
    wprintLib( "Dr.Web sendmail filter shutdown OK." );
    return retcode;
}

/* ----------------------------------------------------------------------- */
static int setuser( void )
{
    struct passwd *account;

    if( UserName[0] != NULLCHAR )
    {
        account = getpwnam( UserName );
        if( !account )
        {
            wprintLib( "non-existing user specified" );
            return DWCLIENT_PARAMETERS;
        }
        if( (setgid(account->pw_gid) < 0)
            ||
            (setuid(account->pw_uid) < 0) )
        {
            wprintLib( "can not change privilegies - %m" );
            return DWCLIENT_SYS_ERR;
        }
    }
    return DWCLIENT_NOERROR;
}

/* ----------------------------------------------------------------------- */
static pid_t daemonize( void )
{
    pid_t          pid;
    int            fd;
    struct rlimit  filelimit;

    signal( SIGTTOU, SIG_IGN );
    signal( SIGTTIN, SIG_IGN );
    signal( SIGTSTP, SIG_IGN );

    pid = fork();
    if( pid == 0 )
    {
        /* create new session without controling terminal */
        if( setsid() < 0 )
        {
            wprintLib( "can not create new session" );
            return -1;
        }
        /* change current dir to allow umount it */
        if( chdir("/") < 0 )
        {
            wprintLib( "can not change directory to root" );
            return -1;
        }
        umask( S_IXUSR | S_IXGRP | S_IRWXO );
        if( getrlimit( RLIMIT_NOFILE, &filelimit ) < 0 )
        {
            wprintLib( "can not get system limits" );
            return -1;
        }
        for( fd = 0; fd < filelimit.rlim_max; ++fd )
        {
            close( fd );
        }
    }

    return pid;
}

/* ----------------------------------------------------------------------- */
static int parseCmdLine( int argc, char ** argv )
{
    int  intVal, Cont;

    if( argc < 2 )
    {
        printf("\nYou should specify parameters.");
        return 0;
    }
    Cont = 1;
    while( Cont )
    {
        switch( getopt( argc, argv, "+wqvrlx:h:t:m:s:i:d:u:n:p:a:f:o:e:g:b:z:y:j:k:") )
        {
            case 'e':
                strncpy( NotifierMail, optarg, sizeof(NotifierMail)-1 );
                break;
            case 'g':
                strncpy( PostmasterMail, optarg, sizeof(PostmasterMail)-1 );
                bPostmasterNotify = 1;
                break;
            case 'o':
                strncpy( UserName, optarg, sizeof(UserName)-1 );
                break;
            case 'f':
                strncpy( ShareDir, optarg, sizeof(ShareDir)-1 );
                break;
            case 'd':
                strncpy( PidFile, optarg, sizeof(PidFile)-1 );
                break;
            case 'a':
                strncpy( ArcDir, optarg, sizeof(ArcDir)-1 );
                break;
            case 'u':
                strncpy( DaemonUnix, optarg, sizeof(DaemonUnix)-1 );
                bDUnix = 1;
                break;
            case 'm':
                strncpy( SendmailUnix, optarg, sizeof(SendmailUnix)-1 );
                bSMUnix = 1;
                break;
            case 'n':
                strncpy( DaemonInet, optarg, sizeof(DaemonInet)-1 );
                bDInet = 1;
                break;
            case 'i':
                strncpy( SendmailInet, optarg, sizeof(SendmailInet)-1 );
                bSMInet = 1;
                break;
            case 'p':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("Error reading port argument (-p option).\n");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    DaemonPort = intVal;
                    bDInet = 1;
                }
                break;
            case 's':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("Error reading port argument (-s option).\n");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    SendmailPort = intVal;
                    bSMInet = 1;
                }
                break;
            case 't':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    printf("Error reading timeout argument (-t timeout).\n");
                    return 0;
                }
                intVal = atoi( optarg );
                if( intVal >= 0 )
                {
                    Timeout = intVal;
                }
                break;
            case 'h':
                switch( optarg[0] )
                {
                    case 'a':
                        scanOptions |= DRWEBD_HEURISTIC_ON;
                        break;
                    case 's':
                        scanOptions |= DRWEBD_SPAM_FILTER;
                        break;
                    default:
                        printf("Invalid argument for option -h\n");
                        break;
                }
                break;
            case 'b':
                if( !set_action( optarg[0], "art", &ActionOnFatalError ) )
                {
                    printf("Invalid argument for option -b\n");
                    return 0;
                }
                break;
            case 'y':
                if( !set_action( optarg[0], "art", &ActionOnError ) )
                {
                    printf("Invalid argument for option -y\n");
                    return 0;
                }
                break;
            case 'z':
                if( !set_action( optarg[0], "cd", &ActionOnSpam ) )
                {
                    printf("Invalid argument for option -z\n");
                    return 0;
                }
                break;
            case 'j':
                if( !set_action( optarg[0], "ar", &ActionOnMailbomb ) )
                {
                    printf("Invalid argument for option -j\n");
                    return 0;
                }
                break;
            case 'k':
                if( !set_action( optarg[0], "ar", &ActionOnSkipObject ) )
                {
                    printf("Invalid argument for option -k\n");
                    return 0;
                }
                break;
            case 'r':
                scanOptions |= DRWEBD_RETURN_VIRUSES;
                break;
            case 'v':
                bCache = 1;
                break;
            case 'q':
                bQuiet = 1;
                break;
            case 'w':
                bDebug = 1;
                break;
            case 'l':
                bLocal = 1;
                break;
            case 'x':
                switch( optarg[0] )
                {
                    case 's':
                        bSenderNotify = 1;
                        break;
                    case 'r':
                        bRcptsNotify = 1;
                        break;
                    default:
                        printf("Invalid argument for option -s\n");
                        break;
                }
                break;
             case -1 :
                Cont = 0;
                break;
            default :
                if( optarg )
                {
                    printf("\nUnknow option <%c>\n",optarg[0]);
                }
                return 0;
        }
    }
    if( NotifierMail[0] == NULLCHAR )
    {
        printf("\nYou should specify e-mail address for notifier.");
        return 0;
    }
    /* !!!
    if( ArcDir[0] == NULLCHAR )
    {
        printf("\nYou should specify directory for infected files.");
        return 0;
    }
    */
    if( ShareDir[0] == NULLCHAR )
    {
        printf("\nYou should specify directory for temporary files.");
        return 0;
    }
    if( bDUnix && bDInet )
    {
        printf("\nYou should define only one address for daemon.");
        return 0;
    }
    if( ((SendmailInet[0] != NULLCHAR) && (SendmailPort == 0))
        ||
        ((SendmailInet[0] == NULLCHAR) && (SendmailPort != 0)) )
    {
        printf("\nYou should define full INET address (name + port) for sendmail.");
        return 0;
    }
    if( (bSMUnix && bSMInet) || (!bSMUnix && !bSMInet) )
    {
        printf("\nYou should define one and only one address for sendmail.");
        return 0;
    }
    return 1;
}
/* ----------------------------------------------------------------------- */
static int usage( void )
{
    printf( "\nUsage: drweb-smf [options]"
            "\n"
            "\n -dFILENAME       Dr.Web daemon PID-file"
            "\n -uSOCKET_FILE    Dr.Web daemon UNIX-socket file"
            "\n -nINETADDRESS    Dr.Web daemon Internet address (IP or domain name)"
            "\n -pPORT           Dr.Web daemon TCP port"
            "\n                  (options -d, -m or a combination of -i and -s are required)"
            "\n -mSOCKET_FILE    Sendmail filter UNIX-socket file"
            "\n -iINETADDRESS    Sendmail filter Internet address (IP or domain name)"
            "\n -sPORT           Sendmail filter TCP port"
            "\n -tTIMEOUT        timeout (in millisec) for wait daemon report"
            "\n -aDIRECTORY      directory for infected messages"
            "\n -l               allow daemon to scan messages as local files"
            "\n -fDIRECTORY      directory where messages are temporarily stored (required)"
            "\n                  (must be accessible for daemon if -l is specified)"
            "\n -ha              use heuristic analyser"
            "\n -hs              use spam analyser"
            "\n -oUSERNAME       user's name"
            "\n -gMAILADDR       admin mail address (may be local) for notifications"
            "\n -eEMAIL          e-mail notifier about viruses,"
            "\n                  public e-mail of postmaster (required)"
            "\n -r               include daemon report in the notification message"
            "\n -xs              send notification to a sender"
            "\n -xr              send notification to a rcpts"
            "\n -bACTION         filter action on fatal errors in filter"
            "\n                  (memory allocation in client or configuration problems)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject (default) | a - accept | t - tempfail }"
            "\n -yACTION         filter action on errors during scanning"
            "\n                  (timeouts or memory alloc problems in daemon for example)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject (default) | a - accept | t - tempfail }"
            "\n -jACTION         filter action on possible mailbomb in message"
            "\n                  (mostly archive with compression > MaxCompressionRation)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject (default) | a - accept }"
            "\n -kACTION         filter action on skipped object in message"
            "\n                  (mostly archives: passworded or part of multivolume)"
            "\n                  Where ACTION one of:"
            "\n                  { r - reject | a - accept (default) }"
            "\n -zACTION         filter action on mails with empty FROM (spam for example):"
            "\n                  Where ACTION one of:"
            "\n                  { c - continue (default) | d - discard }"
            "\n -v               use cached ip-address for daemon (resolved at start)"
            "\n -w               debug mode"
            "\n -q               quiet mode"
            "\n" );
    return DWCLIENT_DEMO_USAGE;
}
