#if defined(__THREAD_BASED_DW)

#ifndef _DWC_MUTEX_C
#define _DWC_MUTEX_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_mutex.c,v 1.4 2001/10/08 15:48:52 cvser Exp $";
#endif

#if defined(__THREAD_BASED_DW)

#include <stdlib.h>
#include <string.h>

#include "dwc_macro.h"
#include "dwc_lib.h"
#include "dwc_err.h"
#include "dwc_mutex.h"

struct DWC_int_mutex_t {
    DWC_string_t      name;
#if defined(__UNIX_DW)
    pthread_mutex_t   handle;
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    HANDLE            handle;
#endif /* __WIN32_DW */
};

/* -------------------------------------------------------------------- */
DWC_mutex_t dwc_mutex_create( DWC_string_t name )
{
    int         error;
    DWC_mutex_t mutex;

    mutex = DWC_Malloc( DWC_mutex_t, sizeof(struct DWC_int_mutex_t) );
    if( !mutex )
    {
        if( name )
            dprintLib( "creating of mutex '%s' failed - no memory.\n", name );
        else
            dprintLib( "creating of anonymous mutex failed - no memory.\n" );
        return NULL;
    }
    if( name )
        mutex->name = strdup( name );
    else
        mutex->name = strdup( "anonymous" );
    if( !mutex->name )
    {
        if( name )
            dprintLib( "creating of mutex '%s' failed - no memory.\n", name );
        else
            dprintLib( "creating of anonymous mutex failed - no memory.\n" );
        DWC_Free( mutex );
        return NULL;
    }
#if defined(__UNIX_DW)
    error = pthread_mutex_init( &mutex->handle, NULL );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    mutex->handle = CreateMutex( NULL, FALSE, mutex->name );
    if( !mutex->handle )
    {
        error = errno;
    }
    else
    {
        error = 0;
    }
#endif /* __WIN32_DW */
    if( error != 0 )
    {
        dprintLib( "creating of mutex '%s' failed - %s.\n", mutex->name, stdlib_error( error ) );
        free( mutex->name );
        DWC_Free( mutex );
        return NULL;
    }
    return mutex;
}

/* -------------------------------------------------------------------- */
DWC_bool_t dwc_mutex_lock( DWC_mutex_t mutex )
{
    int         error;

    PARAM_ASSERT( mutex );

#if defined(__UNIX_DW)
    error = pthread_mutex_lock( &mutex->handle );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( WaitForSingleObject( mutex, INFINITE ) == WAIT_OBJECT_0 )
    {
        error = 0;
    }
    else
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error )
    {
        dprintLib( "locking of mutex '%s' failed - %s.\n", mutex->name, stdlib_error( error ) );
    }
    return ( !error ? TRUE : FALSE );
}

/* -------------------------------------------------------------------- */
DWC_bool_t dwc_mutex_unlock( DWC_mutex_t mutex )
{
    int         error;

    PARAM_ASSERT( mutex );

#if defined(__UNIX_DW)
    error = pthread_mutex_unlock( &mutex->handle );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( ReleaseMutex( mutex->handle ) )
    {
        error = 0;
    }
    else
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error )
    {
        dprintLib( "unlocking of mutex '%s' failed - %s.\n", mutex->name, stdlib_error( error ) );
    }
    return ( !error ? TRUE : FALSE );
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dwc_mutex_destroy( DWC_mutex_t mutex )
{
    DWC_error_code_t    retcode;
    int                 error;

    PARAM_ASSERT( mutex );

#if defined(__UNIX_DW)
    error = pthread_mutex_destroy( &mutex->handle );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( CloseHandle( mutex->handle ) )
    {
        error = 0;
    }
    else
    {
        error = errno;
    }
#endif /* __WIN32_DW */
    if( error != 0 )
    {
        dprintLib( "destroying of mutex '%s' failed - %s.\n", mutex->name, stdlib_error( error ) );
        retcode = DWC_COMMONERROR;
    }
    else
    {
        retcode = DWC_NOERROR;
    }
    free( mutex->name );
    DWC_Free( mutex );
    return retcode;
}

#endif /* __THREAD_BASED_DW */

#endif /* _DWC_MUTEX_C */

#endif /* __THREAD_BASED_DW */
