#if defined(__THREAD_BASED_DW)

#ifndef _DWC_THREAD_C
#define _DWC_THREAD_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_thread.c,v 1.5 2001/10/08 15:48:52 cvser Exp $";
#endif

#ifdef __UNIX_DW
#include <signal.h>
#endif /* __UNIX_DW */
#ifdef __WIN32_DW
#endif /* __WIN32_DW */

#include "dwc_macro.h"
#include "dwc_err.h"
#include "dwc_lib.h"
#include "dwc_thread.h"

struct DWC_int_thread_t {
    DWC_thread_func_t func;
    DWC_thread_arg_t  arg;
#if defined(__UNIX_DW)
    pthread_t         handle;
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    HANDLE            handle;
    DWORD             id;
#endif /* __WIN32_DW */
};
typedef struct DWC_int_thread_t*   DWC_thread_t;

#if defined(__UNIX_DW)
/* -------------------------------------------------------------------- */
void* UnixPthread( void* thread )
{
    DWC_thread_func_t routine;
    DWC_thread_arg_t  argument;
    sigset_t          sigmask;
    int               retcode;

    PARAM_ASSERT( thread );

    routine = ((DWC_thread_t)thread)->func;
    argument = ((DWC_thread_t)thread)->arg;
    DWC_Free( thread );
    sigemptyset( &sigmask );
    sigfillset( &sigmask );
    retcode = pthread_sigmask( SIG_BLOCK, &sigmask, NULL );
    if( retcode != 0 )
    {
        dprintLib( "UnixPthreadRunner: can not block signals for new thread.\n" );
        pthread_exit( NULL );
    }
    (*routine)( argument );
    pthread_exit( NULL );
    return NULL; /* never execute */
}
#endif /* __UNIX_DW */

#if defined(__WIN32_DW)
/* -------------------------------------------------------------------- */
DWORD WINAPI WinThread( LPVOID thread )
{
    DWC_thread_func_t routine;
    DWC_thread_arg_t  argument;

    PARAM_ASSERT( thread );

    routine = ((DWC_thread_t)thread)->func;
    argument = ((DWC_thread_t)thread)->arg;
    DWC_Free( thread );
    (*routine)( argument );
    ExitThread( 0 );
    return 0; /* never execute */
}
#endif /* __WIN32_DW */

/* -------------------------------------------------------------------- */
DWC_error_code_t dwc_thread_create( DWC_thread_func_t func, DWC_thread_arg_t arg )
{
    int             error;
    DWC_thread_t    thread;
#if defined(__UNIX_DW)
    pthread_attr_t  attr;
#endif /* __UNIX_DW */

    PARAM_ASSERT( func );

    thread = DWC_Malloc( DWC_thread_t, sizeof(struct DWC_int_thread_t) );
    if( !thread )
    {
        dprintLib( "Thread creating failed - no memory.\n" );
        return DWC_NOMEMORY;
    }
    thread->func = func;
    thread->arg = arg;
#if defined(__UNIX_DW)
    if( (error = pthread_attr_init( &attr ) == 0)
        &&
        (error = pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED ) == 0) )
    {
        error = pthread_create( &thread->handle, &attr, UnixPthread, (void*) thread );
    }
    pthread_attr_destroy( &attr );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    thread->handle = CreateThread( NULL, 16384, WinThread, (LPVOID) thread, 0, &thread->id );
    if( !thread->handle )
    {
        error = errno;
    }
    else
    {
        error = 0;
    }
    SetThreadPriority( thread->handle, THREAD_PRIORITY_TIME_CRITICAL );
#endif /* __WIN32_DW */
    if( error != 0 )
    {
        dprintLib( "Thread creating failed - %s.\n", stdlib_error( error ) );
        DWC_Free( thread );
        return DWC_COMMONERROR;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dwc_thread_get_id( DWC_thread_id_t* id )
{
    PARAM_ASSERT( id );

#if defined(__UNIX_DW)
    id->id = pthread_self();
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    id->id = GetCurrentThreadId();
#endif /* __WIN32_DW */
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_bool_t dwc_thread_isself( DWC_thread_id_t* id )
{
#if defined(__UNIX_DW)
    return ( id->id == pthread_self() ? TRUE : FALSE );
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    return ( id->id == GetCurrentThreadId() ? TRUE : FALSE );
#endif /* __WIN32_DW */
}

#endif /* _DWC_THREAD_H */

#endif /* __THREAD_BASED_DW */
