#!/usr/bin/perl
#
# $Id: drweb_exim.pl,v 1.1 2001/10/15 11:04:12 cvser Exp $
#
#########################################################
# Acknowledgements:
# Alexey G. Misurenko <mag@caravan.ru>
#    - basic version was ported from Exim integration script
#      that accessible at http://noc.caravan.ru/metro/exim
# Micheal J. Tokarev <mjt@corpit.ru>
#    - for improve design
#    - for security advises

#########################################################
# Settings #
############

# set actions for different situations (Pass, Block)
my $ACTION_ON_MAILBOMB="Block"; # Object with big compression ration found
my $ACTION_ON_SKIP="Pass";      # Object was skipped to check (passworded archive for example)
my $ACTION_ON_ERROR="Block";    # I/O errors, no memory or timeouts, create files and etc

# Enable to save messages that cannot be checked (Yes, No)
# DONT RECOMMENDED use SAVE_UNSCANNED=Yes and ACTION_ON_ERROR="Block" together
my $SAVE_UNSCANNED="No";

# Use spam filter in daemon (Yes,No - casesensitive)
my $SPAM="No";

# Name (domain or IP) of machine where DrWeb daemon runs
#my $SERVER='127.0.0.1';

# DrWeb daemon port on server
#my $PORT='3000';

# Unix socket for DrWeb daemon (if used)
my $SOCKETFILE="__INSTALL_PATH__/run/drwebd.skt";

# E-mail which daemon writes reports from
my $DAEMON_EMAIL="DrWeb-DAEMON\@your.domain";

# Postmaster's e-mail
my $POSTMASTER_EMAIL="postmaster\@your.domain";

# Mail address for alerts and problems reports (can be local)
my $ADMIN_EMAIL="root";

# Mail address for fails (can be local)
my $FAIL_EMAIL="root";

# DONT use for _DIR directory with " " (whitespace) in name

# Directory for save message before check
my $TMP_DIR="__INSTALL_PATH__/spool";

# Use <localscan> mode (Yes,No - casesensitive)
# (if "Yes" then $TMP_DIR should be readable (writeable with EVAL key) for DrWeb Daemon User)
my $LOCAL_MODE="Yes";

# Directory for infected messages
my $INFECTED_DIR="__INSTALL_PATH__/infected";

# Directory for messages that was failed to scan
my $FAIL_DIR="__INSTALL_PATH__/fails";

# Directory where DrWeb is installed
my $DRWEB_DIR="__INSTALL_PATH__";

# Command for postfix wrapper for sendmail
my $EXIM_CMD="/usr/exim/bin/exim";

# System commands
my $MKTEMP_CMD="/bin/mktemp -q";
my $MV_CMD="/bin/mv -f";
my $CP_CMD="/bin/cp -f";

# Timeout for scan operations (in millisecs, 0 for infinite)
my $TIMEOUT='180000';

# Enable (Yes) or disable (Other) logging
my $LOG_ENABLE="No";

# Path to logfile 
my $LOG_FILE="__INSTALL_PATH__/log/exim_filter.log";

# Send virus notification to Sender (Yes,No - case sensitive)
my $NOTIFY_SENDER="Yes";

# Send virus notification to Recipients (Yes,No - case sensitive)
my $NOTIFY_RCPTS="No";

#########################################################
# Variables and constants

my $MAIL_CMD="$EXIM_CMD -bm -i -oMr drweb-scanned -f";

# Construct command line for call drwebd client
my $SCANNER="${DRWEB_DIR}/clients/drwebdc/drwebdc";
my $TIMED;
my $SOCKET_OPT;
if ($TIMEOUT == 0)
{
    $TIMED="";
}
else
{
    $TIMED="-t$TIMEOUT";
}
if ($SOCKETFILE ne "")
{
    $SOCKET_OPT="-u$SOCKETFILE";
}
else
{
    $SOCKET_OPT="-n$SERVER -p$PORT";
}
$SCAN_OPT="-h -rv";
if ($LOCAL_MODE eq "Yes")
{
    $SCAN_OPT.=" -x";
}
if ($SPAM eq "Yes")
{
    $SCAN_OPT.=" -m";
}
my $SCANNER_CMD="$SCANNER $SOCKET_OPT $TIMED $SCAN_OPT -f";

# Exit codes from <sysexits.h>
my $EX_OK=0;
my $EX_USAGE=64;
my $EX_UNAVAILABLE=69;
my $EX_TEMPFAIL=75;

# DrWeb exit status bitmasks
my $DRW_OK=0;                # 0x00
my $DRW_VIRUSFOUND=1;        # 0x01
my $DRW_SKIPOBJECT=2;        # 0x02
my $DRW_MAILBOMB=4;          # 0x04
my $DRW_DAEMONERROR=8;       # 0x08
my $DRW_CLIENTERROR=16;      # 0x10
my $DRW_EVALKEY=32;          # 0x20
my $DRW_SPAM=64;             # 0x40
my $DRW_USAGE=128;           # 0x80

# Some variables
my $StatusMsg="";
my $Line;
my $ExitStatus;

#########################################################
# Start code here
if ($LOG_ENABLE eq "Yes")
{
    open (LOGFILE, ">>$LOG_FILE");
}
else
{
    open (LOGFILE, "/dev/null");
}

# Extract arguments
my $Sender=shift(@ARGV);
if ($Sender eq "") 
{
    $Sender = "\"<>\"";
}
my @Recipient=@ARGV;
# Flush arguments
$#ARGV=0;
pop @ARGV;

logging ("process message: from=$Sender to=@Recipient");

# create temp file
my $TMP_FILE = `$MKTEMP_CMD ${TMP_DIR}/msg.XXXXXX`;
my $Status=($? >> 8);
if ($Status)
{
    $TMP_FILE = "";
    tempfail ("cannot create tempfile in ${TMP_DIR} (error $Status).");
}
# Delete \n at end of filename
# $TMP_FILE =~ s/\n$//;
chomp $TMP_FILE;

open (OUTFILE,">$TMP_FILE") or tempfail ("cannot open $TMP_FILE");
while ($Line = <>)
{
    if (! print OUTFILE $Line) 
    {
        close (OUTFILE);
        tempfail ("cannot save mail to $TMP_FILE");
    }
}
close (OUTFILE) or tempfail ("cannot close $TMP_FILE");

logging ("execute: $SCANNER_CMD $TMP_FILE");
$StatusMsg = `$SCANNER_CMD $TMP_FILE`;
$Status = ($? >> 8);
logging ("daemon client exit code $Status and report: $StatusMsg");
if ($Status == $DRW_OK or $Status == $DRW_EVALKEY)
{
    #virus not found        
    logging ("message $TMP_FILE is ok");
    system ("$MAIL_CMD -i -bm -f $Sender @Recipient < $TMP_FILE") == 0 
        or tempfail ("cannot re-inject message to postfix system");
    $ExitStatus = $EX_OK;
}
else
{
    if (($Status & 0xFFFF) == 0xFFFF)
    {
        tempfail ("daemon client does not exist");
    }
    elsif (($Status & $DRW_VIRUSFOUND) != 0)
    {
        logging ("message $TMP_FILE infected");

        # create archive file
        my $ARC_FILE = `$MKTEMP_CMD ${INFECTED_DIR}/msg.XXXXXX`;
        my $Status=($? >> 8);
        if ($Status)
        {
            tempfail ("cannot create archive file in $INFECTED_DIR");
        }
        # Delete \n at end of filename
        # $ARC_FILE =~ s/\n$//;
        chomp $ARC_FILE;

        # Move infected msg to archive
        system ("$MV_CMD $TMP_FILE $ARC_FILE");
        my $Status=($? >> 8);
        if ($Status)
        {
            unlink ($ARC_FILE);
            tempfail ("cannot move to archive file $ARC_FILE");
        }

        my @VirusInfo = split(/\n/, $StatusMsg);
            
        # mail to Postmaster
        if( open (OUTFILE,"| $MAIL_CMD -i -bm -f $DAEMON_EMAIL $ADMIN_EMAIL") )
        {
            print OUTFILE "From: Antivirus Service <$DAEMON_EMAIL>\n";
            print OUTFILE "To: Postmaster <$ADMIN_EMAIL>\n";
            print OUTFILE "Subject: VIRUS ALERT !\n\n";
            print OUTFILE "Dear Postmaster,\n";
            print OUTFILE "A message that $Sender sent to @Recipient has not been delivered.\n";
            print OUTFILE "Reason: found a virus(es).\n";
            print OUTFILE "Diagnostic:\n";
            foreach $Virus (@VirusInfo)
            {
                print OUTFILE " \t$Virus\n";
            }
            print OUTFILE "An original message has been stored in archive file:\n";
            print OUTFILE "$ARC_FILE\n";
            print OUTFILE "---\n";
            print OUTFILE "\tAntivirus service provided by DrWeb Daemon\n";
            close (OUTFILE) or logging ("cannot send alert to Postmaster");
        }
        else
        {
            logging ("cannot send alert to Postmaster");
        }

        if ($NOTIFY_SENDER eq "Yes")
        {
            # mail to Sender
            if( open (OUTFILE,"| ${MAIL_CMD} -i -bm -f $DAEMON_EMAIL $Sender") )
            {
                print OUTFILE "From: Antivirus Service <$DAEMON_EMAIL>\n";
                print OUTFILE "To: $Sender\n";
                print OUTFILE "Subject: Mail delivery failed (virus detected)\n\n";
                print OUTFILE "This message was created automatically by antivirus filter.\n\n";
                print OUTFILE "A message that you sent to @Recipient has not been delivered.\n";
                print OUTFILE "Reason: found a virus(es).\n";
                print OUTFILE "Diagnostic:\n";
                foreach $Virus (@VirusInfo)
                {
                    print OUTFILE " \t$Virus\n";
                }
                print OUTFILE "An original message has been stored in archive file:\n";
                print OUTFILE "$ARC_FILE\n";
                print OUTFILE "In order to receive original message, please send request to $POSTMASTER_EMAIL, refering to the above filename.\n\n";
                print OUTFILE "---\n";
                print OUTFILE "\tAntivirus service provided by DrWeb Daemon (http://www.drweb.ru)\n";
                close (OUTFILE) or logging ("cannot send alert to $Sender");
            }
            else
            {
                logging ("cannot send alert to $Sender");
            }
        }
        
        if ($NOTIFY_RCPTS eq "Yes")
        {
            # mail to Rcpts
            if( open (OUTFILE,"| ${MAIL_CMD} -i -bm -f $DAEMON_EMAIL @Recipient") )
            {
                print OUTFILE "From: Antivirus Service <$DAEMON_EMAIL>\n";
                print OUTFILE "Subject: Mail delivery failed (virus detected)\n\n";
                print OUTFILE "This message was created automatically by antivirus filter.\n\n";
                print OUTFILE "A message that $Sender sent to you has not been delivered.\n";
                print OUTFILE "Reason: found a virus(es).\n";
                print OUTFILE "Diagnostic:\n";
                foreach $Virus (@VirusInfo)
                {
                    print OUTFILE " \t$Virus\n";
                }
                print OUTFILE "An original message has been stored in archive file:\n";
                print OUTFILE "$ARC_FILE\n";
                print OUTFILE "In order to receive original message, please send request to $POSTMASTER_EMAIL, refering to the above filename.\n\n";
                print OUTFILE "---\n";
                print OUTFILE "\tAntivirus service provided by DrWeb Daemon (http://www.drweb.ru)\n";
                close (OUTFILE) or logging ("cannot send alert to $Sender");
            }
            else
            {
                logging ("cannot send alert to @Recipient");
            }
        }
        close (LOGFILE);
        exit ($EX_OK);
    }
    elsif ((($Status & $DRW_SPAM)) != 0)
    {
        logging ("message $TMP_FILE looks like a spam");
        
        # create archive file
        my $ARC_FILE = `$MKTEMP_CMD ${INFECTED_DIR}/msg.XXXXXX`;
        my $Status=($? >> 8);
        if ($Status)
        {
            tempfail ("cannot create archive file in $INFECTED_DIR");
        }
        # Delete \n at end of filename
        # $ARC_FILE =~ s/\n$//;
        chomp $ARC_FILE;

        # Move infected msg to archive
        system ("$MV_CMD $TMP_FILE $ARC_FILE");
        my $Status=($? >> 8);
        if ($Status)
        {
            unlink ($ARC_FILE);
            tempfail ("cannot move to archive file $ARC_FILE");
        }

        logging ("message $TEMP_FILE dropped and moved to $ARC_FILE");
    
        $ExitStatus = $EX_OK;
    }
    elsif (($Status & $DRW_MAILBOMB) != 0)
    {
        logging ("message $TMP_FILE contains an object(s) that looks like a mailbomb");
        if ($ACTION_ON_MAILBOMB eq "Pass")
        {
            system ("$MAIL_CMD -i -bm -f $Sender @Recipient < $TMP_FILE") == 0
                or tempfail ("cannot re-inject message to postfix system");
            $ExitStatus = $EX_OK;
        }
        else
        {
            print STDOUT "This report was created automatically by antivirus filter.\n\n";
            print STDOUT "A message that you sent to @Recipient contains an object(s) which supposed to be a mailbomb.\n";
            print STDOUT "Please contact with postmaster: $POSTMASTER_EMAIL\n";
            print STDOUT "---\n";
            print STDOUT "\tAntivirus service provided by DrWeb Daemon (http://www.drweb.ru)";
            $ExitStatus = $EX_UNAVAILABLE;
        }
    }
    elsif (($Status & $DRW_SKIPOBJECT) != 0)
    {
        logging ("message $TMP_FILE contains skipped object(s)");
        if ($ACTION_ON_SKIP eq "Pass")
        {
            system ("$MAIL_CMD -i -bm -f $Sender @Recipient < $TMP_FILE") == 0
                or tempfail ("cannot re-inject message to postfix system");
            $ExitStatus = $EX_OK;
        }
        else
        {
            print STDOUT "This report was created automatically by antivirus filter.\n\n";
            print STDOUT "A message that you sent to @Recipient contains an object(s) which cannot be checked (i.e. password protected archive).\n";
            print STDOUT "Please contact with postmaster: $POSTMASTER_EMAIL\n";
            print STDOUT "---\n";
            print STDOUT "\tAntivirus service provided by DrWeb Daemon (http://www.drweb.ru)\n";
            $ExitStatus = $EX_UNAVAILABLE;
        }
    }
    else
    {
        logging ("message $TMP_FILE was not checked: errors were occured during scan");
        # Problem with daemon or client
        my $Who;
        if (($Status & $DRW_CLIENTERROR) != 0)
        {
            $Who = "client";
        }
        else
        {
            $Who = "daemon";
        }
        my $ARC_MSG = "";
        if ($SAVE_UNSCANNED eq "Yes")
        {
            # create archive file
            my $ARC_FILE = `$MKTEMP_CMD ${FAIL_DIR}/msg.XXXXXX`;
            my $Status=($? >> 8);
            if ($Status)
            {
                logging ("cannot create archive file in $FAIL_DIR");
            }
            # Delete \n at end of filename
            # $ARC_FILE =~ s/\n$//;
            chomp $ARC_FILE;

            # Move infected msg to archive
            system ("$CP_CMD $TMP_FILE $ARC_FILE");
            my $Status=($? >> 8);
            if ($Status)
            {
                unlink ($ARC_FILE);
                logging ("cannot move to archive file $ARC_FILE");
            }
            else
            {
                $ARC_MSG = "An original message has been stored in archive file:\n$ARC_FILE\n"
            }
        }
        if ( open (OUTFILE,"| $MAIL_CMD -i -bm -f $DAEMON_EMAIL $ADMIN_EMAIL") )
        {
            print OUTFILE "From: Antivirus Service <$DAEMON_EMAIL>\n";
            print OUTFILE "To: Postmaster <$ADMIN_EMAIL>\n";
            print OUTFILE "Subject: Antivirus Software Failure\n\n";
            print OUTFILE "Dear Postmaster,\n";
            print OUTFILE "Message that $Sender sent to @Recipient has been failed to scan.\n";
            print OUTFILE "Diagnostic: possible problem with $Who (error code - $Status).\n";
            print OUTFILE "$ARC_MSG";
            print OUTFILE "---\n";
            print OUTFILE "\tAntivirus service provided by DrWeb Daemon\n";
            close (OUTFILE) or logging ("cannot send warning to $ADMIN_EMAIL");
        }
        else
        {
            logging ("cannot send warning to $ADMIN_EMAIL");
        }
        if ($ACTION_ON_ERROR eq "Pass")
        {
            system ("$MAIL_CMD -i -bm -f $Sender @Recipient < $TMP_FILE") == 0
                or tempfail ("cannot re-inject message to postfix system");
            $ExitStatus = $EX_OK;
        }
        else
        {
            $ExitStatus = $EX_TEMPFAIL;
        }
    }
}
# Exit, delete temp-file before
close (LOGFILE);
unlink ($TMP_FILE);
exit ($ExitStatus);

#########################################################
# Subroutines

###########
# LOGGING #
###########
sub logging
{
    if ($LOG_ENABLE eq "Yes")
    {
        my $time=localtime;
        print LOGFILE "$time filter[$$] - @_\n";
    }
}

############
# TEMPFAIL #
############
sub tempfail
{
    logging ("problem: @_");    
    if ( open (OUTFILE,"| $MAIL_CMD -i -bm -f $DAEMON_EMAIL $FAIL_EMAIL") )
    {
         print OUTFILE "From: Antivirus Service <$DAEMON_EMAIL>\n";
         print OUTFILE "To: Postmaster <$FAIL_EMAIL>\n";
         print OUTFILE "Subject: Antivirus Software Failure\n\n";
         print OUTFILE "Dear Postmaster,\n";
         print OUTFILE "Diagnostic: @_\n";
         print OUTFILE "---\n";
         print OUTFILE "\tAntivirus service provided by DrWeb Daemon\n";
         close (OUTFILE) or logging ("cannot send fail report to $FAIL_EMAIL");
    }
    else
    {
        logging ("cannot send fail report to $FAIL_EMAIL");
    }    
    close (LOGFILE);
    if ($TMP_FILE ne "")
    {
        unlink ($TMP_FILE);
    }
    exit ($EX_TEMPFAIL);
}
