#!/bin/bash
#
# $Id: drweb_exim.sh,v 1.9 2001/10/15 11:04:12 cvser Exp $
#

#########################################################
# Settings #
############

# set actions for different situations (Pass, Block)
ACTION_ON_MAILBOMB="Block" # Object with big compression ration found
ACTION_ON_SKIP="Pass"      # Object was skipped to check (passworded archive for example)
ACTION_ON_ERROR="Block"    # I/O errors, no memory or timeouts

# Enable to save messages that cannot be checked (Yes, No)
SAVE_UNSCANNED="No"

# Use spam filter in daemon (Yes,No - casesensitive)
SPAM="No"

# Name (domain or IP) machine there DrWeb daemon running
#SERVER="127.0.0.1"

# DrWeb daemon port on server 
#PORT=3000

# Unix socket for DrWeb daemon (if used)
SOCKETFILE="__INSTALL_PATH__/run/drwebd.skt"

# E-Mail from that daemon writes reports
DAEMON_EMAIL="DrWeb-DAEMON@your.domain"

# E-Mail that placed to reports 
POSTMASTER_EMAIL="postmaster@your.domain"

# If in your exim configure file contain non-empty option
# never_users, dont put that users in next two options
# Mail address for alerts and problems reports (can be local)
ADMIN_EMAIL="root"

# Mail address for temporary fails (can be local)
FAIL_EMAIL="root"

# DONT use for _DIR directory with " " (whitespace) in name

# Directory for save message before check
TMP_DIR="__INSTALL_PATH__/spool"

# Use <localscan> mode (Yes,No - casesensitive)
# (if "Yes" then $TMP_DIR should be readable (writeable with EVAL key) for DrWeb Daemon User)
LOCAL_MODE="Yes"

# Directory for infected messages
INFECTED_DIR="__INSTALL_PATH__/infected"

# Directory for messages that was failed to scan
FAIL_DIR="__INSTALL_PATH__/fails"

# Directory where DrWeb installed
DRWEB_DIR="__INSTALL_PATH__"

# System commands
EXIM_CMD="/usr/exim/bin/exim"
DATE_CMD="/bin/date"
MKTEMP_CMD="/bin/mktemp -q"
ECHO_CMD="/bin/echo"
CAT_CMD="/bin/cat"
CP_CMD="/bin/cp -f"
MV_CMD="/bin/mv -f"
RM_CMD="/bin/rm -f"

# Timeout for scan operations (in millisecs, 0 for infinite)
TIMEOUT=180000

# Enable (Yes) or disable (Other) logging
LOG_ENABLE="No"

# Path to logfile 
LOG_FILE="__INSTALL_PATH__/exim_filter.log"

# Send virus notification to Sender (Yes,No - case sensitive)
NOTIFY_SENDER="Yes"

# Send virus notification to Recipients (Yes,No - case sensitive)
NOTIFY_RCPTS="No"

#########################################################
# Variables and constants

MAIL_CMD="$EXIM_CMD -bm -i -oMr drweb-scanned -f"

# Construct command line for call drwebd client
SCANNER="$DRWEB_DIR/clients/drwebdc/drwebdc"
if [ $TIMEOUT -eq 0 ]; then
    TIMED=""
else
    TIMED="-t$TIMEOUT"    
fi
if [ "x"$SOCKETFILE != "x" ] ; then
    # For Unix socket
    SOCKET_OPT="-u$SOCKETFILE"
else
    # For TCP socket
    SOCKET_OPT="-n$SERVER -p$PORT"
fi
SCAN_OPT="-h -rv"
if [ $LOCAL_MODE = "Yes" ] ; then
    SCAN_OPT="$SCAN_OPT -x"
fi
if [ $SPAM = "Yes" ] ; then
    SCAN_OPT="$SCAN_OPT -m"
fi
SCANNER_CMD="$SCANNER $SOCKET_OPT $TIMED $SCAN_OPT -f"

# Exit codes from <sysexits.h>
EX_OK=0
EX_USAGE=64
EX_UNAVAILABLE=69
EX_TEMPFAIL=75

# DrWeb exit status bitmasks
DRW_OK=0                # 0x00
DRW_VIRUSFOUND=1        # 0x01
DRW_SKIPOBJECT=2        # 0x02
DRW_MAILBOMB=4          # 0x04
DRW_DAEMONERROR=8       # 0x08
DRW_CLIENTERROR=16      # 0x10
DRW_EVALKEY=32          # 0x20
DRW_SPAM=64             # 0x40
DRW_USAGE=128           # 0x80

######################
# Logging subroutine #
######################
if [ $LOG_ENABLE == "Yes" ] ; then
   log() 
   {
       $ECHO_CMD "`$DATE_CMD` filter[$$] - $*" >> $LOG_FILE
   }
else
   log() 
   {
       :
   }
fi

#######################
# Tempfail subroutine #
#######################

tempfail() 
{
    log "problem: $@";    
    $ECHO_CMD "From: $DAEMON_EMAIL
To: $FAIL_EMAIL
Subject: Antivirus Software Failure

Dear Postmaster,
Diagnostic: $@
---
   Antivirus service provided by Dr.Web Daemon
" | $MAIL_CMD $DAEMON_EMAIL $FAIL_EMAIL

    if [ $? -ne 0 ] ; then
        log "cannot send tempfail report to $FAIL_EMAIL" 
    fi        
    if [ "x$TEMP_FILE" != "x" ]; then
        $RM_CMD $TEMP_FILE
    fi
    exit $EX_TEMPFAIL
}

##############
# Start code #
##############

# Check environment
if [ ! -d $TMP_DIR ] ; then
    tempfail "$TMP_DIR does not exist"
fi    

if [ ! -x $SCANNER ] ; then
    tempfail "executables $SCANNER does not exist (or permission denied) - you should compile deamon client"
fi    

if [ ! -x $EXIM_CMD ] ; then
    tempfail "executables $EXIM_CMD does not exist (or permission denied)"
fi    

# Parse arguments
Sender="$1"
if [ "x$Sender" = "x" ] ; then
    Sender="\"<>\""   
fi
shift 1 # now $@ contain a recipients of the message

log "process message[$MESSAGE_ID]: from=$Sender to=$@"

# Create tempfile (you need mktemp)
TEMP_FILE=`$MKTEMP_CMD $TMP_DIR/msg.XXXXXX`
err=$?
if [ $err -ne 0 ] ; then
    tempfail "cannot create tmp-file in $TMP_DIR ($err)"
fi

# Clean up when aborting by signal
#trap "$RM_CMD $TEMP_FILE; exit $EX_TEMPFAIL" 2 3 6 15
trap "tempfail \"aborting by signal\""  2 3 6 15

# Save to tempfile.
$CAT_CMD > $TEMP_FILE 
if [ $? -ne 0 ] ; then
    log "cannot save mail to file"
    $RM_CMD $TEMP_FILE
    exit $EX_TEMPFAIL   
fi    

# Call Dr.Web.
EXIT_VIRUS=`$SCANNER_CMD $TEMP_FILE`
EXIT_STATUS=$?

log "DrWebD client exited with code $EXIT_STATUS"

if [ $EXIT_STATUS -eq 126 ] ; then
    tempfail "drwebdc - permission denied"
fi

if [ $EXIT_STATUS -eq 127 ] ; then
    tempfail "drwebdc executable not found"
fi

if [ "x$EXIT_VIRUS" != "x" ] ; then
    log "report: $EXIT_VIRUS"
fi

if [ $EXIT_STATUS -eq $DRW_OK -o $EXIT_STATUS -eq $DRW_EVALKEY ] ; then
    # ok
    log "message $TEMP_FILE is ok."

    # Send normal message
    $MAIL_CMD "$Sender" "$@" < $TEMP_FILE
    if [ $? -ne 0 ] ; then
        tempfail "cannot reinject message $TEMP_FILE to mail system"
    else
        EXIT_CODE=$EX_OK
    fi    
else
    # non-zero exit code
    if [ $(($EXIT_STATUS & $DRW_VIRUSFOUND)) -ne 0 ] ; then
        # infected
        log "message $TEMP_FILE infected." 
    
        # Create filename for archive file
        ARC_FILE=`$MKTEMP_CMD $INFECTED_DIR/msg.XXXXXX`
        if [ $? -ne 0 ] ; then
            tempfail "cannot create file in $INFECTED_DIR"
        fi

        # Rename infected message to archive message file
        $MV_CMD $TEMP_FILE $ARC_FILE
        if [ $? -ne 0 ] ; then
            $RM_CMD $ARC_FILE
            tempfail "cannot move infected message $TEMP_FILE to $ARC_FILE"
        fi                
       
        # Generate and send virus alert to admin
        $ECHO_CMD "From: $DAEMON_EMAIL
To: $ADMIN_EMAIL
Subject: Mail delivery failed (virus detected)

Dear Postmaster,
A message that $Sender sent has not been delivered.
Reason: found a virus(es).
Diagnostic:
    $EXIT_VIRUS
An original message has been stored in archive file:
$ARC_FILE
---
   Antivirus service provided by DrWeb Daemon 
" | $MAIL_CMD $DAEMON_EMAIL $ADMIN_EMAIL
	    
        if [ $? -ne 0 ] ; then
            log "cannot send alert to $ADMIN_EMAIL"
        fi        
        
        if [ $NOTIFY_RCPTS = "Yes" ] ; then
            # Generate and send virus alert to rcpts
            $ECHO_CMD "From: $DAEMON_EMAIL
Subject: Mail delivery failed (virus detected)

Dear Recipient,
A message that $Sender sent to you has not been delivered.
Reason: found a virus(es).
Diagnostic:
    $EXIT_VIRUS
An original message has been stored in archive file:
$ARC_FILE
In order to receive original message, please send request 
to $POSTMASTER_EMAIL, refering to the above filename.
---
   Antivirus service provided by DrWeb Daemon (http://www.drweb.ru)
" | $MAIL_CMD $DAEMON_EMAIL "$@"
	    
            if [ $? -ne 0 ] ; then
                log "cannot send alert to rcpts : $@"
            fi        
        fi
        
        if [ $NOTIFY_SENDER = "Yes" ] ; then
            # Generate and send virus alert to sender
            $ECHO_CMD "From: $DAEMON_EMAIL
Subject: Mail delivery failed (virus detected)

Dear User,
A message that you sent to $@ has not been delivered.
Reason: found a virus(es).
Diagnostic:
    $EXIT_VIRUS
An original message has been stored in archive file:
$ARC_FILE
In order to receive original message, please send request 
to $POSTMASTER_EMAIL, refering to the above filename.
---
   Antivirus service provided by DrWeb Daemon (http://www.drweb.ru)
" | $MAIL_CMD $DAEMON_EMAIL "$Sender"
	    
            if [ $? -ne 0 ] ; then
                log "cannot send alert to $Sender"
            fi        
        fi

        exit $EX_OK
        
    elif [ $(($EXIT_STATUS & $DRW_SPAM)) -ne 0 ] ; then
        # possible spam
        log "message $TEMP_FILE looks like a spam"
        
        # Create filename for archive file
        ARC_FILE=`$MKTEMP_CMD $INFECTED_DIR/msg.XXXXXX`
        if [ $? -ne 0 ] ; then
            tempfail "cannot create file in $INFECTED_DIR"
        fi

        # Rename infected message to archive message file
        $MV_CMD $TEMP_FILE $ARC_FILE
        if [ $? -ne 0 ] ; then
            $RM_CMD $ARC_FILE
            tempfail "cannot move infected message $TEMP_FILE to $ARC_FILE"
        fi                

        log "message $TEMP_FILE dropped and moved to $ARC_FILE."
    
        EXIT_CODE=$EX_OK
        
    elif [ $(($EXIT_STATUS & $DRW_MAILBOMB)) -ne 0 ] ; then
        # possible mail bomb    
        log "message $TEMP_FILE contains an object that looks like a mailbomb."
    
        if [ $ACTION_ON_MAILBOMB == "Pass" ] ; then
            $MAIL_CMD "$Sender" "$@" < $TEMP_FILE
            if [ $? -ne 0 ] ; then
                tempfail "cannot reinject message $TEMP_FILE to mail system" 
            fi
            EXIT_CODE=$EX_OK
        else
            log "message blocked"
            $ECHO_CMD "This report was created automatically by antivirus filter.
            
A message that you sent has not been delivered.
Reason: a message contains an object(s) which supposed to be a mailbomb.
Please contact with postmaster: 
$POSTMASTER_EMAIL
--- 
   Antivirus service provided by DrWeb Daemon (http://www.drweb.ru)" 
            EXIT_CODE=$EX_UNAVAILABLE
        fi
        
    elif [ $(($EXIT_STATUS & $DRW_SKIPOBJECT)) -ne 0 ] ; then
        # skipped object
        log "message $TEMP_FILE contain skipped object."
    
        if [ $ACTION_ON_SKIP == "Pass" ] ; then
            $MAIL_CMD "$Sender" "$@" < $TEMP_FILE
            if [ $? -ne 0 ] ; then
                tempfail "cannot reinject message $TEMP_FILE to mail system" 
            fi
            EXIT_CODE=$EX_OK
        else
            log "message blocked"
            $ECHO_CMD "This report was created automatically by antivirus filter.
            
A message that you sent has not been delivered.
Reason: a message contains an object(s) which cannot be checked (i.e. password protected archive).
Please contact with postmaster: 
$POSTMASTER_EMAIL
---
   Antivirus service provided by DrWeb Daemon (http://www.drweb.ru)"
            EXIT_CODE=$EX_UNAVAILABLE
        fi	
    else
        # errors (daemon, client usage)
        log "message $TEMP_FILE was not checked - some errors during scan."
        
        # Create filename for archive file
        if [ $SAVE_UNSCANNED == "Yes" ]; then
            ARC_FILE=`$MKTEMP_CMD $FAIL_DIR/msg.XXXXXX`
            if [ $? -ne 0 ] ; then
                log "cannot create file in $FAIL_DIR"
            else    
                # Copy unscanned message to archive
                $CP_CMD $TEMP_FILE $ARC_FILE
                if [ $? -ne 0 ] ; then
                    $RM_CMD $ARC_FILE
                    log "cannot move failed message $TEMP_FILE to $ARC_FILE"
                else
                    ARC_MSG="An original message has been stored in archive file:
$ARC_FILE"
                fi                
            fi
        fi
        
        # Generate and send report
        if [ $(($EXIT_STATUS & $DRW_CLIENTERROR)) -ne 0 ] ; then
            who="client"
        else
            who="daemon"
        fi
        $ECHO_CMD "From: $DAEMON_EMAIL
To: $ADMIN_EMAIL
Subject: Antivirus Software Failure

Dear Postmaster,
A message that $Sender sent has been failed to scan.
Diagnostic: possible problem with $who (error code $EXIT_STATUS)
$ARC_MSG
---
   Antivirus service provided by DrWeb Daemon 
" | $MAIL_CMD $DAEMON_EMAIL $ADMIN_EMAIL

        if [ $? -ne 0 ] ; then
            tempfail "cannot send alert to $ADMIN_EMAIL" 
        fi        
        
        if [ $ACTION_ON_ERROR == "Pass" ] ; then
            $MAIL_CMD "$Sender" "$@" < $TEMP_FILE
            if [ $? -ne 0 ] ; then
                tempfail "cannot reinject message $TEMP_FILE to mail system" 
            fi
            EXIT_CODE=$EX_OK
        else
            log "message temporary failed"
            EXIT_CODE=$EX_TEMPFAIL
        fi	
    fi # non-zero exitcodes
fi
# common exit
if [ "x$TEMP_FILE" != "x" ] ; then
    $RM_CMD $TEMP_FILE
fi    
exit $EXIT_CODE

