#!/usr/bin/perl
#
# $Id: drweb_qmail.pl,v 1.16-fix 2001/10/19 12:11:24 cvser Exp $
#
#########################################################
# Acknowledgements:
# MYT <myt@chat.ru>
#  - for explore a possible security hole with (") in filename
#

#########################################################
# Settings #
############

# set actions for different situations (Pass, Block)
my $ACTION_ON_MAILBOMB="Block"; # Object with big compression ration found
my $ACTION_ON_SKIP="Pass";      # Object was skipped to check (passworded archive for example)
my $ACTION_ON_ERROR="Block";    # I/O errors, no memory or timeouts, create files and etc

# Use spam filter in daemon (Yes,No - casesensitive)
my $SPAM="No";

# Name (domain or IP) of machine where DrWeb daemon runs
#my $SERVER='127.0.0.1';

# DrWeb daemon port on server
#my $PORT=3000;

# Unix socket for DrWeb daemon (if used)
my $SOCKETFILE="__INSTALL_PATH__/run/drwebd.skt";

# Directory where DrWeb is installed
my $DRWEB_DIR="__INSTALL_PATH__";

# Timeout for scan operations (in millisecs, 0 for infinite timeout)
my $TIMEOUT=180000;

# Use <localscan> mode (Yes,No - casesensitive)
# (if "Yes" then qmail-scanner spool dir should be readable (writeable with EVAL key) for DrWeb Daemon User)
my $LOCAL_MODE="Yes";

#########################################################
# Init code #
#############

# Construct command line for call drwebd client
my $SCANNER="${DRWEB_DIR}/clients/drwebdc/drwebdc";
my $TIMED;
my $SOCKET_OPT;
if ($TIMEOUT == 0)
{
    $TIMED="";
}
else
{
    $TIMED="-t$TIMEOUT";
}
if ($SOCKETFILE ne "")
{
    $SOCKET_OPT="-u$SOCKETFILE";
}
else
{
    $SOCKET_OPT="-n$SERVER -p$PORT";
}
$SCAN_OPT="-h -rv";
if ($LOCAL_MODE eq "Yes")
{
    $SCAN_OPT.=" -x";
}
if ($SPAM eq "Yes")
{
    $SCAN_OPT.=" -m";
}
my $SCANNER_CMD="$SCANNER $SOCKET_OPT $TIMED $SCAN_OPT -f";

# DrWeb exit status bitmasks
my $DRW_OK=0;                # 0x00
my $DRW_VIRUSFOUND=1;        # 0x01
my $DRW_SKIPOBJECT=2;        # 0x02
my $DRW_MAILBOMB=4;          # 0x04
my $DRW_DAEMONERROR=8;       # 0x08
my $DRW_CLIENTERROR=16;      # 0x10
my $DRW_EVALKEY=32;          # 0x20
my $DRW_USAGE=128;           # 0x80

# DrWeb_Qmail.pl return codes
my $DEX_OK=0;
my $DEX_VIRUS=1;
my $DEX_MAILBOMB=2;
my $DEX_SKIP=3;
my $DEX_ERROR=4;
my $DEX_PASSUNCHECKED=10;

#########################################################
# Main code #
#############
my $drweb_code = $DRW_OK;
my $drweb_output = "";

$drweb_code = DrWebScanDir(shift(@ARGV));

if( ($drweb_code & $DRW_VIRUSFOUND) != 0 )
{
    chomp $drweb_output;
    print "$drweb_output";
    exit( $DEX_VIRUS );
}
elsif( ($drweb_code & $DRW_MAILBOMB) != 0 )
{
    if( $ACTION_ON_MAILBOMB eq "Pass" )
    {
        print "object with exceeded compression ratio";
        exit( $DEX_PASSUNCHECKED );
    }
    exit( $DEX_MAILBOMB );
}
elsif( ($drweb_code & $DRW_SKIPOBJECT) != 0 )
{
    if( $ACTION_ON_SKIP eq "Pass" )
    {
        print "skipped objects";
        exit( $DEX_PASSUNCHECKED );
    }
    exit( $DEX_SKIP );
}
elsif( $drweb_code != $DRW_OK and $drweb_code != $DRW_EVALKEY )
{
    my $who;
    if( ($drweb_code & $DRW_DAEMONERROR) != 0 )
    {
        $who = "daemon";
    }
    else
    {
        $who = "client";
    }
    print "possible problem with $who";
    if( $ACTION_ON_ERROR eq "Pass" )
    {
        exit( $DEX_PASSUNCHECKED );
    }
    exit( $DEX_ERROR );
}
exit( $DEX_OK );

#########################################################
# Traverse directory and scanning #
###################################

sub DrWebScanDir
{
  my $dir = $_[0];
    
  opendir( D, $dir ) || return $DRW_CLIENTERROR;
  # remove parent and current dir-entries
  my @files = grep { ! /^\.$/ && !/^\.\.$/ && (-f "$dir/$_" || -d "$dir/$_") } readdir( D );
  closedir( D ) || return $DRW_CLIENTERROR;

  foreach( @files )
  {
      my $de = "$dir/$_";
      # remove dangerous symbols
      $de =~ s/([\$\`\"\\])/\\$1/gs;
      if( -d $de )
      {
          DrWebScanDir( $de );
      }
      else
      {
          $drweb_output .= `$SCANNER_CMD "$de" 2>&1`;
          $drweb_code |= ($? >> 8);
      }
  }
  return $drweb_code;
}
